from typing import Union
import pandas as pd
import numpy as np
try:
    import ot
except ImportError as error:
    from logging import warning
    warning("{}: ot_distance will be unavailable. To install, run:\n"
            "pip install 'aif360[OptimalTransport]'".format(error))
from sklearn.preprocessing import LabelEncoder

def _normalize(distribution1, distribution2):
    """
    Transform distributions to pleasure form, that is their sums are equal to 1,
    and in case if there is negative values, increase all values with absolute value of smallest number.

    Args:
        distribution1 (numpy array): nontreated distribution
        distribution2 (numpy array): nontreated distribution
    """
    if np.minimum(np.min(distribution1), np.min(distribution2)) < 0:
        extra = -np.minimum(np.min(distribution1), np.min(distribution2))
        distribution1 += extra
        distribution2 += extra

    total_of_distribution1 = np.sum(distribution1)
    if total_of_distribution1 != 0:
        distribution1 /= total_of_distribution1
    total_of_distribution2 = np.sum(distribution2)
    if total_of_distribution2 != 0:
        distribution2 /= total_of_distribution2

def _transform(ground_truth, classifier, cost_matrix=None):
    """
    Transform given distributions from pandas type to numpy arrays, and _normalize them.
    Rearanges distributions, with totall data allocated of one.
    Generates matrix distance with respect to (ground_truth[i] - classifier[j])^2.

    Args:
        ground_truth (series): ground truth (correct) target values
        classifier (series,  dataframe, optional): pandas series estimated targets
            as returned by a model for binary, continuous and ordinal modes.

    Returns:
        initial_distribution, which is an processed ground_truth (numpy array)
        required_distribution, which is an processed classifier (numpy array)
        matrix_distance, which stores the distances between the cells of distributions (2d numpy array)
    """
    initial_distribution = ground_truth.to_numpy().astype(float)
    required_distribution = classifier.to_numpy().astype(float)

    _normalize(initial_distribution, required_distribution)

    if cost_matrix is not None:
        matrix_distance = cost_matrix
    else:
        matrix_distance = np.array([abs(i - required_distribution) for i in initial_distribution], dtype=float)
    return initial_distribution, required_distribution, matrix_distance

def _evaluate(
        ground_truth: pd.Series,
        classifier: pd.Series,
        prot_attr: pd.Series=None,
        num_iters=1e5,
        cost_matrix: np.ndarray=None,
        **kwargs):
    """calculate Wasserstein distance between groups defined by `prot_attr` in `ground_truth` and `classifier`.

    Args:
        ground_truth (pd.Series, str): ground truth (correct) target value
        classifier (pd.Series): estimated target values
        prot_attr (pd.Series, str): pandas series of sensitive attribute values
        num_iters (int, optional): number of iterations (random restarts). Should be positive.

    Returns:
        ot.emd2 (float, dict): Earth mover's distance or dictionary of optimal transports for each of option of classifier
    """

    # Calculate just the EMD between ground_truth and classifier
    if prot_attr is None:
        initial_distribution, required_distribution, matrix_distance = _transform(ground_truth, classifier, cost_matrix)
        return ot.emd2(a=initial_distribution, b=required_distribution, M=matrix_distance, numItermax=num_iters)

    if not ground_truth.nunique() == 2:
        raise ValueError(f"Expected to have exactly 2 target values, got {ground_truth.nunique()}.")

    # Calculate EMD between ground truth distribution and distribution of each group
    emds = {}
    for sa_val in sorted(prot_attr.unique()):
        initial_distribution = ground_truth[prot_attr == sa_val]
        required_distribution = classifier[prot_attr == sa_val]
        initial_distribution, required_distribution, matrix_distance = _transform(initial_distribution, required_distribution, cost_matrix)
        emds[sa_val] = ot.emd2(a=initial_distribution, b=required_distribution, M=matrix_distance, numItermax=num_iters)

    return emds


def ot_distance(
    ground_truth: pd.Series,
    classifier: Union[pd.Series, pd.DataFrame],
    prot_attr: pd.Series = None,
    favorable_value: Union[str, float] = None,
    scoring: str = "Wasserstein1",
    num_iters: int = 1e5,
    penalty: float = 1e-17,
    mode: str = "binary",
    cost_matrix: np.ndarray=None,
    **kwargs,
):
    """Normalize and calculate Wasserstein distance between groups defined by `prot_attr` in `ground_truth` and `classifier`.

    Args:
        ground_truth (pd.Series, str): ground truth (correct) target values.
        classifier (pd.Series, pd.DataFrame, str): estimated target values.
            If `mode` is nominal, must be a dataframe with columns containing predictions for each nominal class.
            If `None`, model is assumed to be a dummy model that predicts the mean of the targets
                or 1/(number of categories) for nominal mode.
        prot_attr (pd.Series, str): sensitive attribute values.
            If `None`, assume all samples belong to the same protected group.
        favorable_value(str, float, optional): Either "high", "low" or a float value if the mode in [binary, ordinal, or continuous].
                If float, value has to be the minimum or the maximum in the ground_truth column.
                Defaults to high if None for these modes.
                Support for float left in to keep the intuition clear in binary classification tasks.
                If `mode` is nominal, favorable values should be one of the unique categories in the ground_truth.
                Defaults to a one-vs-all scan if None for nominal mode.
        scoring (str or class): only 'Wasserstein1'
        num_iters (int, optional): number of iterations (random restarts) for EMD. Should be positive.
        penalty (float, optional): penalty term. Should be positive. The penalty term as with any regularization parameter
            may need to be tuned for a particular use case. The higher the penalty, the higher the influence of entropy regualizer.
        mode: one of ['binary', 'continuous', 'nominal', 'ordinal']. Defaults to binary.
                In nominal mode, up to 10 categories are supported by default.
                To increase this, pass in keyword argument max_nominal = integer value.
        cost_matrix (np.ndarray): cost matrix for the Wasserstein distance. Defaults to absolute difference between samples.

    Returns:
        ot.emd2 (float, dict): Earth mover's distance or dictionary of optimal transports for each of option of classifier

    Raises:
        ValueError: if `mode` is 'binary' but `ground_truth` contains less than 1 or more than 2 unique values.
    """

    # Assert correct mode passed
    if mode not in ['binary', 'continuous', 'nominal', 'ordinal']:
        raise ValueError(f"Expected one of {['binary', 'continuous', 'nominal', 'ordinal']}, got {mode}.")

    # Assert correct types passed to ground_truth, classifier and prot_attr
    if not isinstance(ground_truth, (pd.Series, str)):
        raise TypeError(f"ground_truth: expected pd.Series or str, got {type(ground_truth)}")
    if classifier is not None:
        if mode in ["binary", "continuous"] and not isinstance(classifier, pd.Series):
            raise TypeError(f"classifier: expected pd.Series for {mode} mode, got {type(classifier)}")
        if mode in ["nominal", "ordinal"] and not isinstance(classifier, pd.DataFrame):
            raise TypeError(f"classifier: expected pd.DataFrame for {mode} mode, got {type(classifier)}")
    if prot_attr is not None and not isinstance(prot_attr, (pd.Series, str)):
        raise TypeError(f"prot_attr: expected pd.Series or str, got {type(prot_attr)}")

    # Assert correct type passed to cost_matrix
    if cost_matrix is not None and not isinstance(cost_matrix, np.ndarray):
        raise TypeError(f"cost_matrix: expected numpy.ndarray, got {type(cost_matrix)}")

    # Assert scoring is "Wasserstein1"
    if not scoring == "Wasserstein1":
        raise ValueError(f"Scoring mode can only be \"Wasserstein1\", got {scoring}")

    grt = ground_truth.copy()

    if classifier is not None:
        cls = classifier.copy()
        if prot_attr is not None:
            cls.index = grt.index
    else:
        cls = None

    if prot_attr is not None:
        sat = prot_attr.copy()
        sat.index = grt.index
    else:
        sat = None

    uniques = list(grt.unique())
    if mode == "binary":
        if len(uniques) > 2:
            raise ValueError(f"Only 2 unique values allowed in ground_truth for binary mode, got {uniques}")

    # Encode variables
    if not np.issubdtype(grt.dtype, np.number):
        grt_encoder = LabelEncoder().fit(grt)
        grt = pd.Series(grt_encoder.transform(grt))

    # Set correct favorable value (this tells us if higher or lower is better)
    min_val, max_val = grt.min(), grt.max()

    if favorable_value == 'high':
        favorable_value = max_val
    elif favorable_value == 'low':
        favorable_value = min_val
    elif favorable_value is None:
        if mode in ["binary", "ordinal", "continuous"]:
            favorable_value = max_val # Default to higher is better
        elif mode == "nominal":
            favorable_value = "flag-all" # Default to scan through all categories

    if favorable_value not in [min_val, max_val, "flag-all", *uniques,]:
        raise ValueError(f"Favorable_value should be high, low, or one of categories {uniques}, got {favorable_value}.")

    if mode == "binary": # Flip ground truth if favorable_value is 0 in binary mode.
        grt = pd.Series(grt == favorable_value, dtype=int)
        if cls is None:
            cls = pd.Series(grt.mean(), index=grt.index)
        emds = _evaluate(grt, cls, sat, num_iters, cost_matrix, **kwargs)

    elif mode == "continuous":
        if cls is None:
            cls = pd.Series(grt.mean(), index=grt.index)
        emds = _evaluate(grt, cls, sat, num_iters,cost_matrix, **kwargs)

    ## TODO: rework ordinal mode to take into account distance between pred and true
    elif mode in  ["nominal", "ordinal"]:
        if cls is None: # Set classifier to 1/(num of categories) for nominal mode
            cls = pd.DataFrame([pd.Series(1 / grt.nunique(), index=grt.index)]*grt.nunique())
        if grt.nunique() != cls.shape[-1]:
            raise ValueError(
                f"classifier must have  a column for each class. Expected shape [:, {grt.nunique()}], got {cls.shape}")
        emds = {}
        for class_label in uniques:
            grt_cl = grt.map({class_label: 1}).fillna(0)
            cls_cl = cls[class_label]
            emds[class_label] = _evaluate(grt_cl, cls_cl, sat, num_iters, cost_matrix, **kwargs)

    return emds



# Copyright 2019 Seth V. Neel, Michael J. Kearns, Aaron L. Roth, Zhiwei Steven Wu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""Functions for manipulating and loading input data."""
import argparse
import numpy as np
import pandas as pd


def setup():
    parser = argparse.ArgumentParser(description='Fairness Data Cleaning')
    parser.add_argument(
        '-n',
        '--name',
        type=str,
        help='name of the to store the new datasets (Required)')
    parser.add_argument('-d',
                        '--dataset',
                        type=str,
                        help='name of the original dataset file (Required)')
    parser.add_argument(
        '-a',
        '--attributes',
        type=str,
        help=
        'name of the file representing which attributes are protected (unprotected = 0, protected = 1, label = 2) (Required)'
    )
    parser.add_argument(
        '-c',
        '--centered',
        default=False,
        action='store_true',
        required=False,
        help='Include this flag to determine whether data should be centered')
    args = parser.parse_args()
    return [args.name, args.dataset, args.attributes, args.centered]


def clean_dataset(dataset, attributes, centered):
    """Clean a dataset, given the filename for the dataset and the filename for the attributes.

    Args:
        :param dataset: Filename for dataset. The dataset should be formatted such that categorical
        variables use one-hot encoding
    and the label should be 0/1
        :param attributes: Filename for the attributes of the dataset. The file should have each column name in a list,
         and under this list should have 0 for an unprotected attribute, 1 for a protected attribute, and 2 for the
          attribute of the label.
        :param centered: boolean flag that determines whether to center the input covariates.
        :return X, X_prime, y: pandas dataframes of attributes, sensitive attributes, labels
    """

    df = pd.read_csv(dataset)
    sens_df = pd.read_csv(attributes)

    ## Get and remove label Y
    y_col = [str(c) for c in sens_df.columns if sens_df[c][0] == 2]
    print('label feature: {}'.format(y_col))
    if (len(y_col) > 1):
        raise ValueError('More than 1 label column used')
    if (len(y_col) < 1):
        raise ValueError('No label column used')

    y = df[y_col[0]]

    ## Do not use labels in rest of data
    X = df.loc[:, df.columns != y_col[0]]
    X = X.loc[:, X.columns != 'Unnamed: 0']
    ## Create X_prime, by getting protected attributes
    sens_cols = [str(c) for c in sens_df.columns if sens_df[c][0] == 1]
    print('sensitive features: {}'.format(sens_cols))
    sens_dict = {c: 1 if c in sens_cols else 0 for c in df.columns}
    X, sens_dict = one_hot_code(X, sens_dict)
    sens_names = [key for key in sens_dict.keys() if sens_dict[key] == 1]
    print(
        'there are {} sensitive features including derivative features'.format(
            len(sens_names)))
    X_prime = X[sens_names]
    if centered:
        X = center(X)
        X_prime = center(X_prime)
    return X, X_prime, y


def center(X):
    for col in X.columns:
        X.loc[:, col] = X.loc[:, col] - np.mean(X.loc[:, col])
    return X


def array_to_tuple(x):
    # have to cast ndarray to hashable type in get_baseline()
    x = tuple([el[0] for el in x]) if x.__class__.__name__ == 'ndarray' else x
    return x


def one_hot_code(df1, sens_dict):
    cols = df1.columns
    for c in cols:
        if isinstance(df1[c][0], str):
            column = df1[c]
            df1 = df1.drop(c, 1)
            unique_values = list(set(column))
            n = len(unique_values)
            if n > 2:
                for i in range(n):
                    col_name = '{}.{}'.format(c, i)
                    col_i = [
                        1 if el == unique_values[i] else 0 for el in column
                    ]
                    df1[col_name] = col_i
                    sens_dict[col_name] = sens_dict[c]
                del sens_dict[c]
            else:
                col_name = c
                col = [1 if el == unique_values[0] else 0 for el in column]
                df1[col_name] = col
    return df1, sens_dict


def extract_df_from_ds(dataset):
    """Extract data frames from Transformer Data set

    Args:
         :param dataset: aif360 dataset

    Returns:
         :return X, X_prime, y: pandas dataframes of attributes, sensitive attributes, labels
    """

    X = pd.DataFrame(dataset.convert_to_dataframe()[0])
    # remove labels
    X = X.drop(columns=dataset.label_names)
    # get sensitive attributes
    X_prime = X[dataset.protected_attribute_names]
    y = tuple(dataset.labels[:, 0])
    return X, X_prime, y


def get_data(dataset):
    # Helper for main method
    """Given name of dataset, load in the three datasets associated from the clean.py file
    :param dataset:
    :return:
    """
    X = pd.read_csv('dataset/' + dataset + '_features.csv')
    X_prime = pd.read_csv('dataset/' + dataset + '_protectedfeatures.csv')
    y = pd.read_csv('dataset/' + dataset + '_labels.csv',
                    names=['index', 'label'])
    y = y['label']
    return X, X_prime, y



# Original work Copyright (c) 2017 Geoff Pleiss
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Modified work Copyright 2018 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import numpy as np
from scipy.optimize import linprog

from aif360.algorithms import Transformer
from aif360.metrics import ClassificationMetric, utils


class EqOddsPostprocessing(Transformer):
    """Equalized odds postprocessing is a post-processing technique that solves
    a linear program to find probabilities with which to change output labels to
    optimize equalized odds [8]_ [9]_.

    References:
        .. [8] M. Hardt, E. Price, and N. Srebro, "Equality of Opportunity in
           Supervised Learning," Conference on Neural Information Processing
           Systems, 2016.
        .. [9] G. Pleiss, M. Raghavan, F. Wu, J. Kleinberg, and
           K. Q. Weinberger, "On Fairness and Calibration," Conference on Neural
           Information Processing Systems, 2017.
    """

    def __init__(self, unprivileged_groups, privileged_groups, seed=None):
        """
        Args:
            unprivileged_groups (list(dict)): Representation for unprivileged
                group.
            privileged_groups (list(dict)): Representation for privileged
                group.
            seed (int, optional): Seed to make `predict` repeatable.
        """
        super(EqOddsPostprocessing, self).__init__(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups,
            seed=seed)

        self.seed = seed
        self.model_params = None
        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups

    def fit(self, dataset_true, dataset_pred):
        """Compute parameters for equalizing odds using true and predicted
        labels.

        Args:
            true_dataset (BinaryLabelDataset): Dataset containing true labels.
            pred_dataset (BinaryLabelDataset): Dataset containing predicted
                labels.

        Returns:
            EqOddsPostprocessing: Returns self.
        """
        metric = ClassificationMetric(dataset_true, dataset_pred,
            unprivileged_groups=self.unprivileged_groups,
            privileged_groups=self.privileged_groups)

        # compute basic statistics
        sbr = metric.base_rate(privileged=True)
        obr = metric.base_rate(privileged=False)

        fpr0 = metric.false_positive_rate(privileged=True)
        fpr1 = metric.false_positive_rate(privileged=False)
        fnr0 = metric.false_negative_rate(privileged=True)
        fnr1 = metric.false_negative_rate(privileged=False)
        tpr0 = metric.true_positive_rate(privileged=True)
        tpr1 = metric.true_positive_rate(privileged=False)
        tnr0 = metric.true_negative_rate(privileged=True)
        tnr1 = metric.true_negative_rate(privileged=False)

        # linear program has 4 decision variables:
        # [Pr[label_tilde = 1 | label_hat = 1, protected_attributes = 0];
        #  Pr[label_tilde = 1 | label_hat = 0, protected_attributes = 0];
        #  Pr[label_tilde = 1 | label_hat = 1, protected_attributes = 1];
        #  Pr[label_tilde = 1 | label_hat = 0, protected_attributes = 1]]
        # Coefficients of the linear objective function to be minimized.
        c = np.array([fpr0 - tpr0, tnr0 - fnr0, fpr1 - tpr1, tnr1 - fnr1])

        # A_ub - 2-D array which, when matrix-multiplied by x, gives the values
        # of the upper-bound inequality constraints at x
        # b_ub - 1-D array of values representing the upper-bound of each
        # inequality constraint (row) in A_ub.
        # Just to keep these between zero and one
        A_ub = np.array([[ 1,  0,  0,  0],
                         [-1,  0,  0,  0],
                         [ 0,  1,  0,  0],
                         [ 0, -1,  0,  0],
                         [ 0,  0,  1,  0],
                         [ 0,  0, -1,  0],
                         [ 0,  0,  0,  1],
                         [ 0,  0,  0, -1]], dtype=np.float64)
        b_ub = np.array([1, 0, 1, 0, 1, 0, 1, 0], dtype=np.float64)

        # Create boolean conditioning vectors for protected groups
        cond_vec_priv = utils.compute_boolean_conditioning_vector(
            dataset_pred.protected_attributes,
            dataset_pred.protected_attribute_names,
            self.privileged_groups)
        cond_vec_unpriv = utils.compute_boolean_conditioning_vector(
            dataset_pred.protected_attributes,
            dataset_pred.protected_attribute_names,
            self.unprivileged_groups)

        sconst = np.ravel(
            dataset_pred.labels[cond_vec_priv] == dataset_pred.favorable_label)
        sflip = np.ravel(
            dataset_pred.labels[cond_vec_priv] == dataset_pred.unfavorable_label)
        oconst = np.ravel(
            dataset_pred.labels[cond_vec_unpriv] == dataset_pred.favorable_label)
        oflip = np.ravel(
            dataset_pred.labels[cond_vec_unpriv] == dataset_pred.unfavorable_label)

        y_true = dataset_true.labels.ravel()

        sm_tn = np.logical_and(sflip,
            y_true[cond_vec_priv] == dataset_true.unfavorable_label,
            ).astype(np.float64)
        sm_fn = np.logical_and(sflip,
            y_true[cond_vec_priv] == dataset_true.favorable_label,
            ).astype(np.float64)
        sm_fp = np.logical_and(sconst,
            y_true[cond_vec_priv] == dataset_true.unfavorable_label,
            ).astype(np.float64)
        sm_tp = np.logical_and(sconst,
            y_true[cond_vec_priv] == dataset_true.favorable_label,
            ).astype(np.float64)

        om_tn = np.logical_and(oflip,
            y_true[cond_vec_unpriv] == dataset_true.unfavorable_label,
            ).astype(np.float64)
        om_fn = np.logical_and(oflip,
            y_true[cond_vec_unpriv] == dataset_true.favorable_label,
            ).astype(np.float64)
        om_fp = np.logical_and(oconst,
            y_true[cond_vec_unpriv] == dataset_true.unfavorable_label,
            ).astype(np.float64)
        om_tp = np.logical_and(oconst,
            y_true[cond_vec_unpriv] == dataset_true.favorable_label,
            ).astype(np.float64)

        # A_eq - 2-D array which, when matrix-multiplied by x,
        # gives the values of the equality constraints at x
        # b_eq - 1-D array of values representing the RHS of each equality
        # constraint (row) in A_eq.
        # Used to impose equality of odds constraint
        A_eq = [[(np.mean(sconst*sm_tp) - np.mean(sflip*sm_tp)) / sbr,
                 (np.mean(sflip*sm_fn) - np.mean(sconst*sm_fn)) / sbr,
                 (np.mean(oflip*om_tp) - np.mean(oconst*om_tp)) / obr,
                 (np.mean(oconst*om_fn) - np.mean(oflip*om_fn)) / obr],
                [(np.mean(sconst*sm_fp) - np.mean(sflip*sm_fp)) / (1-sbr),
                 (np.mean(sflip*sm_tn) - np.mean(sconst*sm_tn)) / (1-sbr),
                 (np.mean(oflip*om_fp) - np.mean(oconst*om_fp)) / (1-obr),
                 (np.mean(oconst*om_tn) - np.mean(oflip*om_tn)) / (1-obr)]]

        b_eq = [(np.mean(oflip*om_tp) + np.mean(oconst*om_fn)) / obr
              - (np.mean(sflip*sm_tp) + np.mean(sconst*sm_fn)) / sbr,
                (np.mean(oflip*om_fp) + np.mean(oconst*om_tn)) / (1-obr)
              - (np.mean(sflip*sm_fp) + np.mean(sconst*sm_tn)) / (1-sbr)]

        # Linear program
        self.model_params = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq)

        return self

    def predict(self, dataset):
        """Perturb the predicted labels to obtain new labels that satisfy
        equalized odds constraints.

        Args:
            dataset (BinaryLabelDataset): Dataset containing labels that needs
                to be transformed.
            dataset (BinaryLabelDataset): Transformed dataset.
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        # Get the model parameters output from fit
        sp2p, sn2p, op2p, on2p = self.model_params.x

        # Create boolean conditioning vectors for protected groups
        cond_vec_priv = utils.compute_boolean_conditioning_vector(
            dataset.protected_attributes, dataset.protected_attribute_names,
            self.privileged_groups)
        cond_vec_unpriv = utils.compute_boolean_conditioning_vector(
            dataset.protected_attributes, dataset.protected_attribute_names,
            self.unprivileged_groups)

        # Randomly flip labels according to the probabilities in model_params
        self_fair_pred = dataset.labels[cond_vec_priv].copy()
        self_pp_indices, _ = np.nonzero(
            dataset.labels[cond_vec_priv] == dataset.favorable_label)
        self_pn_indices, _ = np.nonzero(
            dataset.labels[cond_vec_priv] == dataset.unfavorable_label)
        np.random.shuffle(self_pp_indices)
        np.random.shuffle(self_pn_indices)

        n2p_indices = self_pn_indices[:int(len(self_pn_indices) * sn2p)]
        self_fair_pred[n2p_indices] = dataset.favorable_label
        p2n_indices = self_pp_indices[:int(len(self_pp_indices) * (1 - sp2p))]
        self_fair_pred[p2n_indices] = dataset.unfavorable_label

        othr_fair_pred = dataset.labels[cond_vec_unpriv].copy()
        othr_pp_indices, _ = np.nonzero(
            dataset.labels[cond_vec_unpriv] == dataset.favorable_label)
        othr_pn_indices, _ = np.nonzero(
            dataset.labels[cond_vec_unpriv] == dataset.unfavorable_label)
        np.random.shuffle(othr_pp_indices)
        np.random.shuffle(othr_pn_indices)

        n2p_indices = othr_pn_indices[:int(len(othr_pn_indices) * on2p)]
        othr_fair_pred[n2p_indices] = dataset.favorable_label
        p2n_indices = othr_pp_indices[:int(len(othr_pp_indices) * (1 - op2p))]
        othr_fair_pred[p2n_indices] = dataset.unfavorable_label

        # Mutated, fairer dataset with new labels
        dataset_new = dataset.copy()

        new_labels = np.zeros_like(dataset.labels, dtype=np.float64)
        new_labels[cond_vec_priv] = self_fair_pred
        new_labels[cond_vec_unpriv] = othr_fair_pred

        dataset_new.labels = new_labels

        return dataset_new

    def fit_predict(self, dataset_true, dataset_pred):
        """fit and predict methods sequentially."""
        return self.fit(dataset_true, dataset_pred).predict(dataset_pred)



import numpy as np
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import accuracy_score
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

from aif360.algorithms.inprocessing import AdversarialDebiasing as OldAdversarialDebiasing
from aif360.sklearn.inprocessing import AdversarialDebiasing


def test_adv_debias_old_reproduce(old_adult):
    """Test that the old AdversarialDebiasing is reproducible."""
    sess = tf.Session()
    old_adv_deb = OldAdversarialDebiasing(unprivileged_groups=[{'sex': 0}],
                                          privileged_groups=[{'sex': 1}],
                                          scope_name='old_classifier',
                                          sess=sess, num_epochs=5, seed=123)
    old_preds = old_adv_deb.fit_predict(old_adult)
    sess.close()
    tf.reset_default_graph()
    sess = tf.Session()
    old_adv_deb2 = OldAdversarialDebiasing(unprivileged_groups=[{'sex': 0}],
                                          privileged_groups=[{'sex': 1}],
                                          scope_name='old_classifier',
                                          sess=sess, num_epochs=5, seed=123)
    old_preds2 = old_adv_deb2.fit_predict(old_adult)
    sess.close()

    assert np.allclose(old_preds.labels, old_preds2.labels)

def test_adv_debias_old(old_adult, new_adult):
    """Test that the predictions of the old and new AdversarialDebiasing match.
    """
    X, y, _ = new_adult
    tf.reset_default_graph()
    sess = tf.Session()
    old_adv_deb = OldAdversarialDebiasing(unprivileged_groups=[{'sex': 0}],
                                          privileged_groups=[{'sex': 1}],
                                          scope_name='old_classifier',
                                          sess=sess, num_epochs=5, seed=123)
    old_preds = old_adv_deb.fit_predict(old_adult)
    sess.close()
    adv_deb = AdversarialDebiasing('sex', num_epochs=5, random_state=123)
    new_preds = adv_deb.fit(X, y).predict(X)
    adv_deb.sess_.close()
    assert np.allclose(old_preds.labels.flatten(), new_preds)

def test_adv_debias_reproduce(new_adult):
    """Test that the new AdversarialDebiasing is reproducible."""
    X, y, _ = new_adult
    adv_deb = AdversarialDebiasing('sex', num_epochs=5, random_state=123)
    new_preds = adv_deb.fit(X, y).predict(X)
    adv_deb.sess_.close()
    new_acc = accuracy_score(y, new_preds)

    adv_deb2 = AdversarialDebiasing('sex', num_epochs=5, random_state=123)
    new_preds = adv_deb2.fit(X, y).predict(X)
    adv_deb.sess_.close()

    assert new_acc == accuracy_score(y, new_preds)

def test_adv_debias_intersection(new_adult):
    """Test that the new AdversarialDebiasing runs with >2 protected groups."""
    X, y, _ = new_adult
    adv_deb = AdversarialDebiasing(scope_name='intersect', num_epochs=5)
    adv_deb.fit(X, y)
    adv_deb.sess_.close()
    assert adv_deb.adversary_logits_.shape[1] == 4

def test_adv_debias_grid(new_adult):
    """Test that the new AdversarialDebiasing works in a grid search (and that
    debiasing results in reduced accuracy).
    """
    X, y, _ = new_adult
    adv_deb = AdversarialDebiasing('sex', num_epochs=10, random_state=123)

    params = {'debias': [True, False]}

    clf = GridSearchCV(adv_deb, params, cv=3)
    clf.fit(X, y)

    clf.best_estimator_.sess_.close()
    assert clf.best_params_ == {'debias': False}



"""This is the helper script for implementing metrics."""
import numpy as np


def compute_boolean_conditioning_vector(X, feature_names, condition=None):
    """Compute the boolean conditioning vector.

    Args:
        X (numpy.ndarray): Dataset features
        feature_names (list): Names of the features.
        condition (list(dict)): Specifies the subset of instances we want to
            use. Format is a list of `dicts` where the keys are `feature_names`
            and the values are values in `X`. Elements in the list are clauses
            joined with OR operators while key-value pairs in each dict are
            joined with AND operators. See examples for more details. If `None`,
            the condition specifies the entire set of instances, `X`.

    Returns:
        numpy.ndarray(bool): Boolean conditioning vector. Shape is `[n]` where
        `n` is `X.shape[0]`. Values are `True` if the corresponding row
        satisfies the `condition` and `False` otherwise.

    Examples:
        >>> condition = [{'sex': 1, 'age': 1}, {'sex': 0}]

        This corresponds to `(sex == 1 AND age == 1) OR (sex == 0)`.
    """
    if condition is None:
        return np.ones(X.shape[0], dtype=bool)

    overall_cond = np.zeros(X.shape[0], dtype=bool)
    for group in condition:
        group_cond = np.ones(X.shape[0], dtype=bool)
        for name, val in group.items():
            index = feature_names.index(name)
            group_cond = np.logical_and(group_cond, X[:, index] == val)
        overall_cond = np.logical_or(overall_cond, group_cond)

    return overall_cond

def compute_num_instances(X, w, feature_names, condition=None):
    """Compute the number of instances, :math:`n`, conditioned on the protected
    attribute(s).

    Args:
        X (numpy.ndarray): Dataset features.
        w (numpy.ndarray): Instance weight vector.
        feature_names (list): Names of the features.
        condition (list(dict)): Same format as
            :func:`compute_boolean_conditioning_vector`.

    Returns:
        int: Number of instances (optionally conditioned).
    """

    # condition if necessary
    cond_vec = compute_boolean_conditioning_vector(X, feature_names, condition)

    return np.sum(w[cond_vec], dtype=np.float64)

def compute_num_pos_neg(X, y, w, feature_names, label, condition=None):
    """Compute the number of positives, :math:`P`, or negatives, :math:`N`,
    optionally conditioned on protected attributes.

    Args:
        X (numpy.ndarray): Dataset features.
        y (numpy.ndarray): Label vector.
        w (numpy.ndarray): Instance weight vector.
        feature_names (list): Names of the features.
        label (float): Value of label (unfavorable/positive or
            unfavorable/negative).
        condition (list(dict)): Same format as
            :func:`compute_boolean_conditioning_vector`.

    Returns:
        int: Number of positives/negatives (optionally conditioned)
    """
    y = y.ravel()
    cond_vec = compute_boolean_conditioning_vector(X, feature_names,
        condition=condition)
    return np.sum(w[np.logical_and(y == label, cond_vec)], dtype=np.float64)

def compute_num_TF_PN(X, y_true, y_pred, w, feature_names, favorable_label,
                      unfavorable_label, condition=None):
    """Compute the number of true/false positives/negatives optionally
    conditioned on protected attributes.

    Args:
        X (numpy.ndarray): Dataset features.
        y_true (numpy.ndarray): True label vector.
        y_pred (numpy.ndarray): Predicted label vector.
        w (numpy.ndarray): Instance weight vector - the true and predicted
            datasets are supposed to have same instance level weights.
        feature_names (list): names of the features.
        favorable_label (float): Value of favorable/positive label.
        unfavorable_label (float): Value of unfavorable/negative label.
        condition (list(dict)): Same format as
            :func:`compute_boolean_conditioning_vector`.

    Returns:
        Number of positives/negatives (optionally conditioned).
    """
    # condition if necessary
    cond_vec = compute_boolean_conditioning_vector(X, feature_names,
        condition=condition)

    # to prevent broadcasts
    y_true = y_true.ravel()
    y_pred = y_pred.ravel()

    y_true_pos = (y_true == favorable_label)
    y_true_neg = (y_true == unfavorable_label)
    y_pred_pos = np.logical_and(y_pred == favorable_label, cond_vec)
    y_pred_neg = np.logical_and(y_pred == unfavorable_label, cond_vec)

    # True/false positives/negatives
    return dict(
        TP=np.sum(w[np.logical_and(y_true_pos, y_pred_pos)], dtype=np.float64),
        FP=np.sum(w[np.logical_and(y_true_neg, y_pred_pos)], dtype=np.float64),
        TN=np.sum(w[np.logical_and(y_true_neg, y_pred_neg)], dtype=np.float64),
        FN=np.sum(w[np.logical_and(y_true_pos, y_pred_neg)], dtype=np.float64)
    )

def compute_num_gen_TF_PN(X, y_true, y_score, w, feature_names, favorable_label,
                    unfavorable_label, condition=None):
    """Compute the number of generalized true/false positives/negatives
    optionally conditioned on protected attributes. Generalized counts are based
    on scores and not on the hard predictions.

    Args:
        X (numpy.ndarray): Dataset features.
        y_true (numpy.ndarray): True label vector.
        y_score (numpy.ndarray): Predicted score vector. Values range from 0 to
            1. 0 implies prediction for unfavorable label and 1 implies
            prediction for favorable label.
        w (numpy.ndarray): Instance weight vector - the true and predicted
            datasets are supposed to have same instance level weights.
        feature_names (list): names of the features.
        favorable_label (float): Value of favorable/positive label.
        unfavorable_label (float): Value of unfavorable/negative label.
        condition (list(dict)): Same format as
            :func:`compute_boolean_conditioning_vector`.

    Returns:
        Number of positives/negatives (optionally conditioned).
    """
    # condition if necessary
    cond_vec = compute_boolean_conditioning_vector(X, feature_names,
        condition=condition)

    # to prevent broadcasts
    y_true = y_true.ravel()
    y_score = y_score.ravel()
    w = w.ravel()

    y_true_pos = np.logical_and(y_true == favorable_label, cond_vec)
    y_true_neg = np.logical_and(y_true == unfavorable_label, cond_vec)

    # Generalized true/false positives/negatives
    return dict(
        GTP=np.sum((w*y_score)[y_true_pos], dtype=np.float64),
        GFP=np.sum((w*y_score)[y_true_neg], dtype=np.float64),
        GTN=np.sum((w*(1.0-y_score))[y_true_neg], dtype=np.float64),
        GFN=np.sum((w*(1.0-y_score))[y_true_pos], dtype=np.float64)
    )

def compute_distance(X_orig, X_distort, X_prot, feature_names, dist_fun,
                     condition=None):
    """Compute the distance element-wise for two sets of vectors.

    Args:
        X_orig (numpy.ndarray): Original features.
        X_distort (numpy.ndarray): Distorted features. Shape must match
            `X_orig`.
        X_prot (numpy.ndarray): Protected attributes (used to compute
            condition). Should be same for both original and distorted.
        feature_names (list): Names of the protected features.
        dist_fun (function): Function which returns the distance (float) between
            two 1-D arrays (e.g. :func:`scipy.spatial.distance.euclidean`).
        condition (list(dict)): Same format as
            :func:`compute_boolean_conditioning_vector`.

    Returns:
        (numpy.ndarray(numpy.float64), numpy.ndarray(bool)):

            * Element-wise distances (1-D).
            * Condition vector (1-D).
    """
    cond_vec = compute_boolean_conditioning_vector(X_prot, feature_names,
        condition=condition)

    num_instances = X_orig[cond_vec].shape[0]
    distance = np.zeros(num_instances, dtype=np.float64)
    for i in range(num_instances):
        distance[i] = dist_fun(X_orig[cond_vec][i], X_distort[cond_vec][i])

    return distance, cond_vec



import numpy as np

from aif360.algorithms.preprocessing.reweighing import Reweighing
from aif360.datasets import AdultDataset

def test_instance_weights():
    ad = AdultDataset(instance_weights_name='fnlwgt', features_to_drop=[])
    privileged_groups = [{'sex': 1}]
    unprivileged_groups = [{'sex': 0}]
    rw = Reweighing(unprivileged_groups=unprivileged_groups,
                    privileged_groups=privileged_groups)
    transf = rw.fit_transform(ad)
    print(transf.instance_weights.sum())
    assert np.isclose(ad.instance_weights.sum(), transf.instance_weights.sum())



from collections import OrderedDict
import json

from aif360.explainers import MetricTextExplainer
from aif360.metrics import BinaryLabelDatasetMetric


class MetricJSONExplainer(MetricTextExplainer):
    """Class for explaining metric values in JSON format.

    These briefly explain what a metric is and/or how it is calculated unless
    it is obvious (e.g. accuracy) and print the value.

    This class contains JSON explanations for all metric values regardless of
    which subclass they appear in. This will raise an error if the metric does
    not apply (e.g. calling `true_positive_rate` if
    `type(metric) == DatasetMetric`).
    """

    def accuracy(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).accuracy(privileged=privileged)
        response = OrderedDict((
            ("metric", "Accuracy"),
            ("message", outcome),
            ("numTruePositives", self.metric.num_true_positives(privileged=privileged)),
            ("numTrueNegatives", self.metric.num_true_negatives(privileged=privileged)),
            ("numPositives", self.metric.num_positives(privileged=privileged)),
            ("numNegatives", self.metric.num_negatives(privileged=privileged)),
            ("description", "Computed as (true positive count + "
                "true negative count)/(positive_count + negative_count)."),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def average_abs_odds_difference(self):
        outcome = super(MetricJSONExplainer, self).average_abs_odds_difference()
        response = OrderedDict((
            ("metric", "Average Absolute Odds Difference"),
            ("message", outcome),
            ("numFalsePositivesUnprivileged", self.metric.num_false_positives(privileged=False)),
            ("numNegativesUnprivileged", self.metric.num_negatives(privileged=False)),
            ("numTruePositivesUnprivileged", self.metric.num_true_positives(privileged=False)),
            ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
            ("numFalsePositivesPrivileged", self.metric.num_false_positives(privileged=True)),
            ("numNegativesPrivileged", self.metric.num_negatives(privileged=True)),
            ("numTruePositivesPrivileged", self.metric.num_true_positives(privileged=True)),
            ("numPositivesPrivileged", self.metric.num_positives(privileged=True)),
            ("description", "Computed as average difference of false positive rate (false positives / actual negatives) and true positive rate (true positives / actual positives) between unprivileged and privileged groups."),
            ("ideal", "The ideal value of this metric is 0.0.  A value of < 0 implies higher benefit for the privileged group and a value > 0 implies higher benefit for the unprivileged group.")
        ))
        return json.dumps(response)

    def average_odds_difference(self):
        outcome = super(MetricJSONExplainer, self).average_odds_difference()
        response = OrderedDict((
            ("metric", "Average Odds Difference"),
            ("message", outcome),
            ("numFalsePositivesUnprivileged", self.metric.num_false_positives(privileged=False)),
            ("numNegativesUnprivileged", self.metric.num_negatives(privileged=False)),
            ("numTruePositivesUnprivileged", self.metric.num_true_positives(privileged=False)),
            ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
            ("numFalsePositivesPrivileged", self.metric.num_false_positives(privileged=True)),
            ("numNegativesPrivileged", self.metric.num_negatives(privileged=True)),
            ("numTruePositivesPrivileged", self.metric.num_true_positives(privileged=True)),
            ("numPositivesPrivileged", self.metric.num_positives(privileged=True)),
            ("description", "Computed as average difference of false positive rate (false positives / negatives) and true positive rate (true positives / positives) between unprivileged and privileged groups."),
            ("ideal", "The ideal value of this metric is 0.  A value of < 0 implies higher benefit for the privileged group and a value > 0 implies higher benefit for the unprivileged group.")
        ))
        return json.dumps(response)

    def between_all_groups_coefficient_of_variation(self):
        outcome = super(MetricJSONExplainer, self).between_all_groups_coefficient_of_variation()
        response = OrderedDict((
            ("metric", "Between All Groups Coefficient Of Variation"),
            ("message", outcome),
            ("description", "Computed as the square root of twice the pairwise entropy between every pair of privileged and underprivileged groups with alpha = 2."),
            ("ideal", "The ideal value of this metric is 0.") #2.0"
        ))
        return json.dumps(response)

    def between_all_groups_generalized_entropy_index(self, alpha=2):
        outcome = super(MetricJSONExplainer, self).between_all_groups_generalized_entropy_index(alpha)
        response = OrderedDict((
            ("metric", "Between All Groups Generalized Entropy Index"),
            ("message", outcome),
            ("description", "Computed as the pairwise entropy between every pair of privileged and underprivileged groups."),
            ("ideal", "The ideal value of this metric is 0.0") #1.0"
        ))
        return json.dumps(response)

    def between_all_groups_theil_index(self):
        outcome = super(MetricJSONExplainer, self).between_all_groups_theil_index()
        response = OrderedDict((
            ("metric", "Between All Groups Theil Index"),
            ("message", outcome),
            ("description", "Computed as the pairwise entropy between every pair of privileged and underprivileged groups with alpha = 1."),
            ("ideal", "The ideal value of this metric is 0.0") #1.0"
        ))
        return json.dumps(response)

    def between_group_coefficient_of_variation(self):
        outcome = super(MetricJSONExplainer, self).between_group_coefficient_of_variation()
        response = OrderedDict((
            ("metric", "Between Group Coefficient Of Variation"),
            ("message", outcome),
            ("description", "Computed as the square root of twice the pairwise entropy between a given pair of privileged and underprivileged groups with alpha = 2."),
            ("ideal", "The ideal value of this metric is 0.0") #2.0"
        ))
        return json.dumps(response)

    def between_group_generalized_entropy_index(self, alpha=2):
        outcome = super(MetricJSONExplainer, self).between_group_generalized_entropy_index(alpha)
        response = OrderedDict((
            ("metric", "Between Group Generalized Entropy Index"),
            ("message", outcome),
            ("description", "Computed as the pairwise entropy between a given pair of privileged and underprivileged groups."),
            ("ideal", "The ideal value of this metric is 0.0") #1.0"
        ))
        return json.dumps(response)

    def between_group_theil_index(self):
        outcome = super(MetricJSONExplainer, self).between_group_theil_index()
        response = OrderedDict((
            ("metric", "Between Group Theil Index"),
            ("message", outcome),
            ("description", "Computed as the pairwise entropy between a given pair of privileged and underprivileged groups with alpha = 1."),
            ("ideal", "The ideal value of this metric is 0.0") #1.0"
        ))
        return json.dumps(response)

    def coefficient_of_variation(self):
        outcome = super(MetricJSONExplainer, self).coefficient_of_variation()
        response = OrderedDict((
            ("metric", "Coefficient Of Variation"),
            ("message", outcome),
            ("description", "Computed as the square root of twice the generalized entropy index with alpha = 2."),
            ("ideal", "The ideal value of this metric is 0.0") #2.0"
        ))
        return json.dumps(response)

    def consistency(self, n_neighbors=5):
        outcome = super(MetricJSONExplainer, self).consistency(n_neighbors)
        response = OrderedDict((
            ("metric", "Consistency"),
            ("message", outcome),
            ("description", "Individual fairness metric from Zemel, Rich, et al. \"Learning fair representations.\", ICML 2013. "
                            "Measures how similar the labels are for similar instances."),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def disparate_impact(self):
        outcome = super(MetricJSONExplainer, self).disparate_impact()
        response = []
        if type(self.metric) is BinaryLabelDatasetMetric:
            response = OrderedDict((
                ("metric", "Disparate Impact"),
                ("message", outcome),
                ("numPositivePredictionsUnprivileged", self.metric.num_positives(privileged=False)),
                ("numUnprivileged", self.metric.num_instances(privileged=False)),
                ("numPositivePredictionsPrivileged", self.metric.num_positives(privileged=True)),
                ("numPrivileged", self.metric.num_instances(privileged=True)),
                ("description", "Computed as the ratio of rate of favorable outcome for the unprivileged group to that of the privileged group."),
                ("ideal", "The ideal value of this metric is 1.0 A value < 1 implies higher benefit for the privileged group and a value >1 implies a higher benefit for the unprivileged group.")
            ))
        else:
            response = OrderedDict((
                ("metric", "Disparate Impact"),
                ("message", outcome),
                ("numPositivePredictionsUnprivileged", self.metric.num_pred_positives(privileged=False)),
                ("numUnprivileged", self.metric.num_instances(privileged=False)),
                ("numPositivePredictionsPrivileged", self.metric.num_pred_positives(privileged=True)),
                ("numPrivileged", self.metric.num_instances(privileged=True)),
                ("description", "Computed as the ratio of likelihood of favorable outcome for the unprivileged group to that of the privileged group."),
                ("ideal", "The ideal value of this metric is 1.0")
            ))
        return json.dumps(response)

    def error_rate(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).error_rate(privileged=privileged)
        response = OrderedDict((
            ("metric", "Error Rate"),
            ("message", outcome),
            ("numTruePositives", self.metric.num_true_positives(privileged=privileged)),
            ("numTrueNegatives", self.metric.num_true_negatives(privileged=privileged)),
            ("numPositives", self.metric.num_positives(privileged=privileged)),
            ("numNegatives", self.metric.num_negatives(privileged=privileged)),
            ("description", "Computed as  (1 -(true positive count + true negative count)/(positive_count + negative_count)). "),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def error_rate_difference(self):
        outcome = super(MetricJSONExplainer, self).error_rate_difference()
        response = OrderedDict((
            ("metric", "Error Rate Difference"),
            ("message", outcome),
            ("numTruePositivesUnprivileged", self.metric.num_true_positives(privileged=False)),
            ("numTrueNegativesUnprivileged", self.metric.num_true_negatives(privileged=False)),
            ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
            ("numNegativesUnprivileged", self.metric.num_negatives(privileged=False)),
            ("numTruePositivesPrivileged", self.metric.num_true_positives(privileged=True)),
            ("numTrueNegativesPrivileged", self.metric.num_true_negatives(privileged=True)),
            ("numPositivePrivileged", self.metric.num_positives(privileged=True)),
            ("numNegativePrivileged", self.metric.num_negatives(privileged=True)),
            ("description", "Error rate = 1 -(true positive count + true negative count)/(positive_count + negative_count). "
                "This metric is computed as the difference of error rates between unprivileged and privileged groups."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def error_rate_ratio(self):
        outcome = super(MetricJSONExplainer, self).error_rate_ratio()
        response = OrderedDict((
            ("metric", "Error Rate Ratio"),
            ("message", outcome),
            ("numTruePositivesUnprivileged", self.metric.num_true_positives(privileged=False)),
            ("numTrueNegativesUnprivileged", self.metric.num_true_negatives(privileged=False)),
            ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
            ("numNegativesUnprivileged", self.metric.num_negatives(privileged=False)),
            ("numTruePositivesPrivileged", self.metric.num_true_positives(privileged=True)),
            ("numTrueNegativesPrivileged", self.metric.num_true_negatives(privileged=True)),
            ("numPositivePrivileged", self.metric.num_positives(privileged=True)),
            ("numNegativePrivileged", self.metric.num_negatives(privileged=True)),
            ("description", "Error rate = 1 -(true positive count + true negative count)/(positive_count + negative_count). "
                "This metric is computed as the ratio of error rates between unprivileged and privileged groups."),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def false_discovery_rate(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).false_discovery_rate(privileged=privileged)
        response = OrderedDict((
            ("metric", "False Discovery Rate"),
            ("message", outcome),
            ("numTruePositives", self.metric.num_true_positives(privileged=privileged)),
            ("numFalsePositives", self.metric.num_false_positives(privileged=privileged)),
            ("description", "Computed as  (false positive count / (true positive count + false positive count))."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_discovery_rate_difference(self):
        outcome = super(MetricJSONExplainer, self).false_discovery_rate_difference()
        response = OrderedDict((
            ("metric", "False Discovery Rate Difference"),
            ("message", outcome),
            ("numTruePositivesUnprivileged", self.metric.num_true_positives(privileged=False)),
            ("numFalsePositivesUnprivileged", self.metric.num_false_positives(privileged=False)),
            ("numTruePositivesPrivileged", self.metric.num_true_positives(privileged=True)),
            ("numFalsePositivesPrivileged", self.metric.num_false_positives(privileged=True)),
            ("description", "False discovery rate is computed as  (false positive count / (true positive count + false positive count)). "
                "This metric is computed as the difference of false discovery rate of unprivileged and privileged instances."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_discovery_rate_ratio(self):
        outcome = super(MetricJSONExplainer, self).false_discovery_rate_ratio()
        response = OrderedDict((
            ("metric", "False Discovery Rate Ratio"),
            ("message", outcome),
            ("numTruePositivesUnprivileged", self.metric.num_true_positives(privileged=False)),
            ("numFalsePositivesUnprivileged", self.metric.num_false_positives(privileged=False)),
            ("numTruePositivesPrivileged", self.metric.num_true_positives(privileged=True)),
            ("numFalsePositivesPrivileged", self.metric.num_false_positives(privileged=True)),
            ("description", "False discovery rate is computed as  (false positive count / (true positive count + false positive count)). "
                "This metric is computed as the ratio of false discovery rate of unprivileged and privileged instances."),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def false_negative_rate(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).false_negative_rate(privileged=privileged)
        response = OrderedDict((
            ("metric", "False Negative Rate"),
            ("message", outcome),
            ("numFalseNegatives", self.metric.num_false_negatives(privileged=privileged)),
            ("numPositives", self.metric.num_positives(privileged=privileged)),
            ("description", "Computed as  (false negagive count / total positive count)."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_negative_rate_difference(self):
        outcome = super(MetricJSONExplainer, self).false_negative_rate_difference()
        response = OrderedDict((
            ("metric", "False Negative Rate Difference"),
            ("message", outcome),
            ("numFalseNegativesUnprivileged", self.metric.num_false_negatives(privileged=False)),
            ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
            ("numFalseNegativesPrivileged", self.metric.num_false_negatives(privileged=True)),
            ("numPositivesPrivileged", self.metric.num_positives(privileged=True)),
            ("description", "False negative rate is computed as  (false negagive count / total positive count). "
                "This metric is computed as the difference of false negative rate between unprivileged and privileged instances."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_negative_rate_ratio(self):
        outcome = super(MetricJSONExplainer, self).false_negative_rate_ratio()
        response = OrderedDict((
            ("metric", "False Negative Rate Ratio"),
            ("message", outcome),
            ("numFalseNegativesUnprivileged", self.metric.num_false_negatives(privileged=False)),
            ("numPositiveaUnprivileged", self.metric.num_positives(privileged=False)),
            ("numFalseNegativesPrivileged", self.metric.num_false_negatives(privileged=True)),
            ("numPositiveaPrivileged", self.metric.num_positives(privileged=True)),
            ("description", "False negative rate is computed as  (false negagive count / total positive count). "
                "This metric is computed as the ratio of false negative rate between unprivileged and privileged instances."),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def false_omission_rate(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).false_omission_rate(privileged=privileged)
        response = OrderedDict((
            ("metric", "False Omission Rate"),
            ("message", outcome),
            ("numTrueNegatives", self.metric.num_true_negatives(privileged=privileged)),
            ("numFalseNegatives", self.metric.num_false_negatives(privileged=privileged)),
            ("description", "Computed as  (false negative count / (true negative count + false negative count))."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_omission_rate_difference(self):
        outcome = super(MetricJSONExplainer, self).falses_omission_rate_difference()
        response = OrderedDict((
            ("metric", "False Omission Rate Difference"),
            ("message", outcome),
            ("numTrueNegativesUnprivileged", self.metric.num_true_negatives(privileged=False)),
            ("numFalseNegativesUnprivileged", self.metric.num_false_negatives(privileged=False)),
            ("numTrueNegativesPrivileged", self.metric.num_true_negatives(privileged=True)),
            ("numFalseNegativesPrivileged", self.metric.num_false_negatives(privileged=True)),
            ("description", "False omission rate is computed as  (false negative count / (true negative count + false negative count)). "
                "This metric is computed as the difference of false omission rate of underprivileged and privileged groups."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_omission_rate_ratio(self):
        outcome = super(MetricJSONExplainer, self).false_omission_rate_ratio()
        response = OrderedDict((
            ("metric", "False Omission Rate Ratio"),
            ("message", outcome),
            ("numTrueNegativesUnprivileged", self.metric.num_true_negatives(privileged=False)),
            ("numFalseNegativesUnprivileged", self.metric.num_false_negatives(privileged=False)),
            ("numTrueNegativesPrivileged", self.metric.num_true_negatives(privileged=True)),
            ("numFalseNegativesPrivileged", self.metric.num_false_negatives(privileged=True)),
            ("description", "False omission rate is computed as  (false negative count / (true negative count + false negative count)). "
                "This metric is computed as the ratio of false omission rate of underprivileged and privileged groups."),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def false_positive_rate(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).false_positive_rate(privileged=privileged)
        response = OrderedDict((
            ("metric", "False Positive Rate"),
            ("message", outcome),
            ("numFalsePositives", self.metric.num_false_positives(privileged=privileged)),
            ("numNegatives", self.metric.num_negatives(privileged=privileged)),
            ("description", "Computed as (false positive count / total negative count)."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_positive_rate_difference(self):
        outcome = super(MetricJSONExplainer, self).false_positive_rate_difference()
        response = OrderedDict((
            ("metric", "False Positive Rate Difference"),
            ("message", outcome),
            ("numFalsePositivesUnprivileged", self.metric.num_false_positives(privileged=False)),
            ("numNegativesUnprivileged", self.metric.num_negatives(privileged=False)),
            ("numPositivesPrivileged", self.metric.num_false_positives(privileged=True)),
            ("numNegativesPrivileged", self.metric.num_negatives(privileged=True)),
            ("description", "False positive rate is computed as (false positive count / total negative count). "
                "This metric is computed as the difference of false positive rates between the unprivileged and privileged groups"),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def false_positive_rate_ratio(self):
        outcome = super(MetricJSONExplainer, self).false_positive_rate_ratio()
        response = OrderedDict((
            ("metric", "False Positive Rate Ratio"),
            ("message", outcome),
            ("numFalsePositivesUnprivileged", self.metric.num_false_positives(privileged=False)),
            ("numNegativesUnprivileged", self.metric.num_negatives(privileged=False)),
            ("numFalsePositivesPrivileged", self.metric.num_false_positives(privileged=True)),
            ("numNegativesPrivileged", self.metric.num_negatives(privileged=True)),
            ("description", "False positive rate is computed as (false positive count / total negative count). "
                "This metric is computed as the ratio of false positive rates between the unprivileged and privileged groups"),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def generalized_entropy_index(self, alpha=2):
        outcome = super(MetricJSONExplainer, self).generalized_entropy_index(alpha=alpha)
        response = OrderedDict((
            ("metric", "Generalized Entropy Index"),
            ("message", outcome),
            ("description", "This metric represents the generalized entropy index measured between the predicted and actual favorable outcomes."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def mean_difference(self):
        outcome = super(MetricJSONExplainer, self).mean_difference()
        response = OrderedDict((
            ("metric", "Mean Difference"),
            ("message", outcome),
            ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
            ("numInstancesUnprivileged", self.metric.num_instances(privileged=False)),
            ("numPositivesPrivileged", self.metric.num_positives(privileged=True)),
            ("numInstancesPrivileged", self.metric.num_instances(privileged=True)),
            ("description", "Computed as the difference of the rate of favorable outcomes received by the unprivileged group to the privileged group."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def negative_predictive_value(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).negative_predictive_value(privileged=privileged)
        response = OrderedDict((
            ("metric", "Negative Predictive Value"),
            ("message", outcome),
            ("numTrueNegatives", self.metric.num_true_negatives(privileged=privileged)),
            ("numFalseNegatives", self.metric.num_false_negatives(privileged=privileged)),
            ("description", "Computed as (number of true negatives / (number of true negatives + number of false negatives))."),
            ("ideal", "The ideal value of this metric is 1.0")
        ))
        return json.dumps(response)

    def num_false_negatives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_false_negatives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of False Negatives"),
            ("message", outcome),
            ("numFalseNegatives", self.metric.num_false_negatives(privileged=privileged)),
            ("description", "Computed as the number of false negative instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def num_false_positives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_false_positives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of False Positives"),
            ("message", outcome),
            ("numFalsePositives", self.metric.num_false_positives(privileged=privileged)),
            ("description", "Computed as the number of false positive instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value of this metric is 0.0")
        ))
        return json.dumps(response)

    def num_instances(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_instances(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of Instances"),
            ("message", outcome),
            ("numInstances", self.metric.num_instances(privileged=privileged)),
            ("description", "Computed as the number of instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is the total number of instances made available")
        ))
        return json.dumps(response)

    def num_negatives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_negatives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of Negatives"),
            ("message", outcome),
            ("numNegatives", self.metric.num_negatives(privileged=privileged)),
            ("description", "Computed as the number of negative instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value of this metric lies in the total number of negative instances made available")
        ))
        return json.dumps(response)

    def num_positives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_positives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of Positives"),
            ("message", outcome),
            ("numPositives", self.metric.num_positives(privileged=privileged)),
            ("description", "Computed as the number of positive instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value of this metric lies in the total number of positive instances made available")
        ))
        return json.dumps(response)

    def num_pred_negatives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_pred_negatives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of Predicted Negatives"),
            ("message", outcome),
            ("numPredNegatives", self.metric.num_pred_negatives(privileged=privileged)),
            ("description", "Computed as the number of predicted negative instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is the total number of negative instances made available")
        ))
        return json.dumps(response)

    def num_pred_positives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_pred_positives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of Predicted Positives"),
            ("message", outcome),
            ("numPredPositives", self.metric.num_pred_positives(privileged=privileged)),
            ("description", "Computed as the number of predicted positive instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is the total number of positive instances made available")
        ))
        return json.dumps(response)

    def num_true_negatives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_true_negatives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of True Negatives"),
            ("message", outcome),
            ("numTrueNegatives", self.metric.num_true_negatives(privileged=privileged)),
            ("description", "Computed as the number of true negative instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is the total number of negative instances made available")
        ))
        return json.dumps(response)

    def num_true_positives(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).num_true_positives(privileged=privileged)
        response = OrderedDict((
            ("metric", "Number Of True Positives"),
            ("message", outcome),
            ("numTruePositives", self.metric.num_true_positives(privileged=privileged)),
            ("description", "Computed as the number of true positive instances for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is the total number of positive instances made available")
        ))
        return json.dumps(response)

    def positive_predictive_value(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).positive_predictive_value(privileged=privileged)
        response = OrderedDict((
            ("metric", "Positive Predictive Value"),
            ("message", outcome),
            ("numTruePositives", self.metric.num_true_positives(privileged=privileged)),
            ("numFalsePositives", self.metric.num_false_positives(privileged=privileged)),
            ("description", "Computed as (true positives / (true positives + false positives)) for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is 1.0")
        ))
        return json.dumps(response)

    def statistical_parity_difference(self):
        outcome = super(MetricJSONExplainer, self).statistical_parity_difference()
        response = []
        if type(self.metric) is BinaryLabelDatasetMetric:
            response = OrderedDict((
                ("metric", "Statistical Parity Difference"),
                ("message", outcome),
                ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
                ("numInstancesUnprivileged", self.metric.num_instances(privileged=False)),
                ("numPositivesPrivileged", self.metric.num_positives(privileged=True)),
                ("numInstancesPrivileged", self.metric.num_instances(privileged=True)),
                ("description", "Computed as the difference of the rate of favorable outcomes received by the unprivileged group to the privileged group."),
                ("ideal", " The ideal value of this metric is 0")
            ))
        else:
            response = OrderedDict((
                ("metric", "Statistical Parity Difference"),
                ("message", outcome),
                ("numPositivesUnprivileged", self.metric.num_pred_positives(privileged=False)),
                ("numInstancesUnprivileged", self.metric.num_instances(privileged=False)),
                ("numPositivesPrivileged", self.metric.num_pred_positives(privileged=True)),
                ("numInstancesPrivileged", self.metric.num_instances(privileged=True)),
                ("description", "Computed as the difference of the rate of favorable outcomes received by the unprivileged group to the privileged group."),
                ("ideal", " The ideal value of this metric is 0")
            ))
        return json.dumps(response)

    def theil_index(self):
        outcome = super(MetricJSONExplainer, self).theil_index()
        response = OrderedDict((
            ("metric", "Theil Index"),
            ("message", outcome),
            ("description", "Computed as the generalized entropy of benefit for all individuals in the dataset, with alpha = 1. It measures the inequality in benefit allocation for individuals."),
            ("ideal", "A value of 0 implies perfect fairness.")
        ))
        return json.dumps(response)

    def true_negative_rate(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).true_negative_rate(privileged=privileged)
        response = OrderedDict((
            ("metric", "True Negative Rate"),
            ("message", outcome),
            ("numTrueNegatives", self.metric.num_true_negatives(privileged=privileged)),
            ("numNegatives", self.metric.num_negatives(privileged=privileged)),
            ("description", "Computed as the ratio of true negatives to the total number of negatives for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is 1.0")
        ))
        return json.dumps(response)

    def true_positive_rate(self, privileged=None):
        outcome = super(MetricJSONExplainer, self).true_positive_rate(privileged=privileged)
        response = OrderedDict((
            ("metric", "True Positive Rate"),
            ("message", outcome),
            ("numTruePositives", self.metric.num_true_positives(privileged=privileged)),
            ("numPositives", self.metric.num_positives(privileged=privileged)),
            ("description", "Computed as the ratio of true positives to the total number of positives for the given (privileged or unprivileged) group."),
            ("ideal", "The ideal value is 1.0")
        ))
        return json.dumps(response)

    def true_positive_rate_difference(self):
        outcome = super(MetricJSONExplainer, self).true_positive_rate_difference()
        response = OrderedDict((
            ("metric", "True Positive Rate Difference"),
            ("message", outcome),
            ("numTruePositivesUnprivileged", self.metric.num_true_positives(privileged=False)),
            ("numPositivesUnprivileged", self.metric.num_positives(privileged=False)),
            ("numTruePositivesPrivileged", self.metric.num_true_positives(privileged=True)),
            ("numPositivesPrivileged", self.metric.num_positives(privileged=True)),
            ("description", "This metric is computed as the difference of true positive rates between the unprivileged and the privileged groups. "
                " The true positive rate is the ratio of true positives to the total number of actual positives for a given group."),
            ("ideal", "The ideal value is 0. A value of < 0 implies higher benefit for the privileged group and a value > 0 implies higher benefit for the unprivileged group.")
        ))
        return json.dumps(response)

    # ============================== ALIASES ===================================
    def equal_opportunity_difference(self):
        return self.true_positive_rate_difference()

    def power(self, privileged=None):
        return self.num_true_positives(privileged=privileged)

    def precision(self, privileged=None):
        return self.positive_predictive_value(privileged=privileged)

    def recall(self, privileged=None):
        return self.true_positive_rate(privileged=privileged)

    def sensitivity(self, privileged=None):
        return self.true_positive_rate(privileged=privileged)

    def specificity(self, privileged=None):
        return self.true_negative_rate(privileged=privileged)



import numpy as np
import pandas as pd
from aif360.metrics.ot_metric import _normalize, _transform, _evaluate
from aif360.sklearn.metrics import ot_distance
from ot import emd2
import unittest
from unittest import TestCase

rng = np.random.default_rng(seed = 42)
s = 100
d1 = rng.normal(loc=0, size=s)
d2 = rng.normal(loc=10, size=s)
sd1 = pd.Series(d1)
sd2 = pd.Series(d2)
_e_ = np.minimum(np.min(d1), np.min(d2))
_d1_= d1
_d2_ = d2
if _e_ < 0:
    _d1_ -= _e_
    _d2_ -= _e_
_d1_ /= np.sum(d1)
_d2_ /= np.sum(d2)
dist = np.array([abs(i - _d2_) for i in _d1_], dtype=float)

class TestNormalizeTransform(TestCase):
    def test_normalize(self):
        # test normalization: must make every value non-negative
        _normalize(d1, d2)
        assert isinstance(d1, np.ndarray), f"_normalize must keep inputs as ndarray, got {type(d1)}"
        assert isinstance(d2, np.ndarray), f"_normalize must keep inputs as ndarray, got {type(d2)}"
        assert np.all(d1 >= 0), "_normalize: negatives present in distribution 1"
        assert np.all(d2 >= 0), "_normalize: negatives present in distribution 2"
        assert abs(np.sum(d1) - 1) < 1e-6, "_normalize: distribution 1 must sum to 1"
        assert abs(np.sum(d2) - 1) < 1e-6, "_normalize: distribution 2 must sum to 1"

    def test_transform(self):
        # check if transform returns np.ndarrays, makes sums equal
        d1_, d2_, dist_ = _transform(sd1, sd2, None)
        assert isinstance(d1_, np.ndarray), f"_transform must return ndarrays, got {type(d1_)}"
        assert isinstance(d2_, np.ndarray), f"_transform must return ndarrays, got {type(d2_)}"
        s1, s2 = np.sum(d1_), np.sum(d2_)
        assert abs(s1 - s2) < 1e-6, f"_transform must return arrays with equal sums, got {s1} and {s2}"
        assert np.all(np.abs(dist - dist_)) < 1e-6, "_transform distance matrix not calculated correctly"

class TestEvaluate():

    def test_evaluate_quantecon(self):
        # check against example in https://python.quantecon.org/opt_transport.html
        # with normalization
        p = pd.Series([50, 100, 150])
        q = pd.Series([25, 115, 60, 30, 70])
        C = np.array([[10, 15, 20, 20, 40], 
                      [20, 40, 15, 30, 30], 
                      [30, 35, 40, 55, 25]])
        expected = 24.083333
        actual = _evaluate(p, q, cost_matrix=C)
        assert abs(expected - actual) < 1e-6

    def test_evaluate_normal(self):
        # check against PyOptimalTransport's EMD2
        a_ = d1/np.sum(d1)
        b_ = d2/np.sum(d2)
        dist = np.array([abs(i - b_) for i in a_], dtype=float)
        expected = emd2(a_, b_, dist)
        actual = _evaluate(sd1, sd2, num_iters = 100000)
        assert abs(expected-actual) < 1e-3, f"EMD must be {expected}, got {actual}"

    # check properties of a metric
    def test_evaluate_same(self):
        # emd(x, x) = 0
        expected = 0
        actual = _evaluate(sd1, sd1, num_iters = 1000)
        assert abs(expected - actual) < 1e-8, f"EMD between two equal distributions must be 0, got {actual}"

    def test_evaluate_symmetry(self):
        a = _evaluate(sd1, sd2, num_iters = 1000)
        b = _evaluate(sd2, sd1, num_iters = 1000)
        assert abs(a - b) < 1e-8, f"EMD must be symmetric, got {a} and {b}"

    def test_evaluate_triangle(self):
        d3 = pd.Series(rng.normal(loc=2, size=s))
        a = _evaluate(sd1, sd2, num_iters = 1000)
        b = _evaluate(sd2, d3, num_iters = 1000)
        c = _evaluate(sd1, d3, num_iters = 1000)
        assert a + b >= c, f"EMD must satisfy triangle inequality"
    
    def test_binary(self):
        p = pd.Series([0,1,0,1])
        q = pd.Series([0.7,0.7,0.3,0.3])
        s = pd.Series([0,0,1,1])
        expected = {sv: _evaluate(p[s==sv], q[s==sv]) for sv in s}
        actual = _evaluate(p, q, s)
        assert expected == actual

class TestOtBiasScan(TestCase):
    def test_scoring_checked(self):
        with self.assertRaises(Exception):
            ot_distance(pd.Series(), pd.Series(), scoring="Bernoulli")
    def test_scan_mode_checked(self):
        with self.assertRaises(Exception):
            ot_distance(pd.Series(), pd.Series(), mode="Wrong")

    def test_classifier_type_checked(self):
        with self.assertRaises(Exception):
            ot_distance(pd.Series(), pd.Series(), mode="nominal")
        with self.assertRaises(Exception):
            ot_distance(pd.Series(), pd.Series(), mode="ordinal")
        with self.assertRaises(Exception):
            ot_distance(pd.Series(), pd.DataFrame(), mode="binary")
        with self.assertRaises(Exception):
            ot_distance(pd.Series(), pd.DataFrame(), mode="continuous")

    def test_binary_nuniques_checked(self):
        with self.assertRaises(Exception):
            ot_distance(pd.Series([1,2,3], pd.Series(), mode='binary'))
    
    def test_cost_matrix_type_checked(self):
        with self.assertRaises(TypeError):
            ot_distance(pd.Series(), pd.Series(), cost_matrix=pd.DataFrame())

    def test_cost_matrix_passed_correctly(self):
        p = pd.Series([50, 100, 150])
        q = pd.Series([25, 115, 60, 30, 70])
        C = np.array([[10, 15, 20, 20, 40], 
                      [20, 40, 15, 30, 30], 
                      [30, 35, 40, 55, 25]])
        expected = _evaluate(p, q, cost_matrix=C)
        actual = ot_distance(p, q, cost_matrix=C, mode="continuous")
        assert expected == actual

    def test_favorable_value_checked(self):
        p = pd.Series([50, 100, 150])
        q = pd.Series([25, 115, 60, 30, 70])
        fav = 4
        with self.assertRaises(ValueError):
            ot_distance(p, q, pos_label=fav)

    def test_nominal_classifier_shape_checked(self):
        p = pd.Series([0,0,1,1])
        q = pd.DataFrame([0.5,0.5,0.5,0.5])
        with self.assertRaises(ValueError):
            ot_distance(p, q, mode='nominal')
    
    def test_ordinal_classifier_shape_checked(self):
        p = pd.Series([0,0,1,1])
        q = pd.DataFrame([0.5,0.5,0.5,0.5])
        with self.assertRaises(ValueError):
            ot_distance(p, q, mode='ordinal')

    def test_nominal(self):
        p = pd.Series([0,0,1,1])
        q = pd.DataFrame([[0.5,0.5],[0.5,0.5],[0.5,0.5],[0.5,0.5]])
        expected = {cl: _evaluate(p, q[cl]) for cl in [0,1]}
        actual = ot_distance(p, q, mode="nominal")
        assert expected == actual


if __name__ == '__main__':
    unittest.main()


import pandas as pd
from sklearn.model_selection import train_test_split
from aif360.datasets import RegressionDataset
from aif360.sklearn.datasets.lawschool_dataset import LSAC_URL

class LawSchoolGPADataset(RegressionDataset):
    """Law School GPA dataset."""

    def __init__(self, dep_var_name='zfygpa',
                 protected_attribute_names=['race', 'gender'],
                 privileged_classes=[['white'], ['male']],
                 instance_weights_name=None,
                 categorical_features=[],
                 na_values=[], custom_preprocessing=None,
                 metadata=None):
        """See :obj:`RegressionDataset` for a description of the arguments."""
        df = pd.read_sas(LSAC_URL, encoding="utf-8")
        df.race = df.race1.where(df.race1.isin(['black', 'white']))
        df.gender = df.gender.fillna("female")
        df = df[["race", "gender", "lsat", "ugpa", "zfygpa"]].dropna()
        df = pd.concat(train_test_split(df, test_size=0.33, random_state=123))

        super(LawSchoolGPADataset, self).__init__(df=df,
            dep_var_name=dep_var_name,
            protected_attribute_names=protected_attribute_names,
            privileged_classes=privileged_classes,
            instance_weights_name=instance_weights_name,
            categorical_features=categorical_features,
            na_values=na_values,
            custom_preprocessing=custom_preprocessing, metadata=metadata)



import numpy as np

from aif360.datasets import AdultDataset
from aif360.metrics import ClassificationMetric
from aif360.algorithms.inprocessing import MetaFairClassifier


protected = 'sex'
ad = AdultDataset(protected_attribute_names=[protected],
                    privileged_classes=[['Male']], categorical_features=[],
                    features_to_keep=['age', 'education-num', 'capital-gain',
                                    'capital-loss', 'hours-per-week'])
test, train = ad.split([16281], shuffle=False)

def test_adult_sr():
    biased_model = MetaFairClassifier(tau=0, sensitive_attr=protected,
                                      type='sr', seed=123).fit(train)
    dataset_bias_test = biased_model.predict(test)

    biased_cm = ClassificationMetric(test, dataset_bias_test,
            unprivileged_groups=[{protected: 0}],
            privileged_groups=[{protected: 1}])
    spd1 = biased_cm.disparate_impact()
    spd1 = min(spd1, 1/spd1)

    debiased_model = MetaFairClassifier(tau=0.9, sensitive_attr=protected,
                                        type='sr', seed=123).fit(train)
    dataset_debiasing_test = debiased_model.predict(test)

    debiased_cm = ClassificationMetric(test, dataset_debiasing_test,
            unprivileged_groups=[{protected: 0}],
            privileged_groups=[{protected: 1}])
    spd2 = debiased_cm.disparate_impact()
    spd2 = min(spd2, 1/spd2)
    assert(spd2 >= spd1)

def test_adult_fdr():
    biased_model = MetaFairClassifier(tau=0, sensitive_attr=protected,
                                      type='fdr', seed=123).fit(train)
    dataset_bias_test = biased_model.predict(test)

    biased_cm = ClassificationMetric(test, dataset_bias_test,
            unprivileged_groups=[{protected: 0}],
            privileged_groups=[{protected: 1}])
    fdr1 = biased_cm.false_discovery_rate_ratio()
    fdr1 = min(fdr1, 1/fdr1)

    debiased_model = MetaFairClassifier(tau=0.9, sensitive_attr=protected,
                                        type='fdr', seed=123).fit(train)
    dataset_debiasing_test = debiased_model.predict(test)

    debiased_cm = ClassificationMetric(test, dataset_debiasing_test,
            unprivileged_groups=[{protected: 0}],
            privileged_groups=[{protected: 1}])
    fdr2 = debiased_cm.false_discovery_rate_ratio()
    fdr2 = min(fdr2, 1/fdr2)
    assert(fdr2 >= fdr1)



import os

import pandas as pd

from aif360.datasets import StandardDataset


class BankDataset(StandardDataset):
    """Bank marketing Dataset.

    See :file:`aif360/data/raw/bank/README.md`.
    """

    def __init__(self, label_name='y', favorable_classes=['yes'],
                 protected_attribute_names=['age'],
                 privileged_classes=[lambda x: x >= 25 and x < 60],
                 instance_weights_name=None,
                 categorical_features=['job', 'marital', 'education', 'default',
                     'housing', 'loan', 'contact', 'month', 'day_of_week',
                     'poutcome'],
                 features_to_keep=[], features_to_drop=[],
                 na_values=["unknown"], custom_preprocessing=None,
                 metadata=None):
        """See :obj:`StandardDataset` for a description of the arguments.

        By default, this code converts the 'age' attribute to a binary value
        where privileged is `25 <= age < 60` and unprivileged is `age < 25` or `age >= 60`
        as suggested in Le Quy, Tai, et al. [1].

        References:
            .. [1] Le Quy, Tai, et al. "A survey on datasets for fairness‐aware machine 
            learning." Wiley Interdisciplinary Reviews: Data Mining and Knowledge 
            Discovery 12.3 (2022): e1452.


        """

        filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
            '..', 'data', 'raw', 'bank', 'bank-additional-full.csv')

        try:
            df = pd.read_csv(filepath, sep=';', na_values=na_values)
        except IOError as err:
            print("IOError: {}".format(err))
            print("To use this class, please download the following file:")
            print("\n\thttps://archive.ics.uci.edu/ml/machine-learning-databases/00222/bank-additional.zip")
            print("\nunzip it and place the files, as-is, in the folder:")
            print("\n\t{}\n".format(os.path.abspath(os.path.join(
               os.path.abspath(__file__), '..', '..', 'data', 'raw', 'bank'))))
            import sys
            sys.exit(1)

        super(BankDataset, self).__init__(df=df, label_name=label_name,
            favorable_classes=favorable_classes,
            protected_attribute_names=protected_attribute_names,
            privileged_classes=privileged_classes,
            instance_weights_name=instance_weights_name,
            categorical_features=categorical_features,
            features_to_keep=features_to_keep,
            features_to_drop=features_to_drop, na_values=na_values,
            custom_preprocessing=custom_preprocessing, metadata=metadata)



from sklearn.linear_model import LogisticRegression

from aif360.datasets import AdultDataset
from aif360.algorithms.postprocessing import EqOddsPostprocessing
from aif360.algorithms.postprocessing import CalibratedEqOddsPostprocessing
from aif360.metrics import ClassificationMetric

train, val, test = AdultDataset().split([0.4, 0.7])
lr = LogisticRegression(solver='lbfgs').fit(train.features, train.labels)

val_pred = val.copy()
val_pred.labels = lr.predict(val.features).reshape((-1, 1))
val_pred.scores = lr.predict_proba(val.features)[:, 1]

pred = test.copy()
pred.labels = lr.predict(test.features).reshape((-1, 1))
pred.scores = lr.predict_proba(test.features)[:, 1]

cm_lr = ClassificationMetric(test, pred, unprivileged_groups=[{'sex': 0}],
                             privileged_groups=[{'sex': 1}])

def test_eqodds():
    eqo = EqOddsPostprocessing(unprivileged_groups=[{'sex': 0}],
                               privileged_groups=[{'sex': 1}], seed=1234567)
    pred_eqo = eqo.fit(val, val_pred).predict(pred)
    cm_eqo = ClassificationMetric(test, pred_eqo,
            unprivileged_groups=[{'sex': 0}], privileged_groups=[{'sex': 1}])
    # accuracy drop should be less than 10% (arbitrary)
    assert (cm_lr.accuracy() - cm_eqo.accuracy()) / cm_lr.accuracy() < 0.1
    # approximately equal odds
    assert cm_eqo.average_abs_odds_difference() < 0.1

def test_caleq():
    ceo = CalibratedEqOddsPostprocessing(cost_constraint='fnr',
            unprivileged_groups=[{'sex': 0}],
            privileged_groups=[{'sex': 1}], seed=1234567)
    pred_ceo = ceo.fit(val, val_pred).predict(pred)

    cm_ceo = ClassificationMetric(test, pred_ceo,
            unprivileged_groups=[{'sex': 0}], privileged_groups=[{'sex': 1}])
    # accuracy drop should be less than 10% (arbitrary)
    assert (cm_lr.accuracy() - cm_ceo.accuracy()) / cm_lr.accuracy() < 0.1
    # approximate GFNR parity
    assert abs(cm_ceo.difference(cm_ceo.generalized_false_negative_rate)) < 0.1



import numpy as np


class ScoringFunction:
    def __init__(self, **kwargs):
        """
        This is an abstract class for Scoring Functions (or expectation-based scan statistics).

        [1] introduces a property of many commonly used log-likelihood ratio scan statistics called
        Additive linear-time subset scanning (ALTSS) that allows for exact of efficient maximization of these
        statistics over all subsets of the data, without requiring an exhaustive search over all subsets and
        allows penalty terms to be included.

        [1] Speakman, S., Somanchi, S., McFowland III, E., & Neill, D. B. (2016). Penalized fast subset scanning.
        Journal of Computational and Graphical Statistics, 25(2), 382-404.
        """
        self.kwargs = kwargs
        self.direction = kwargs.get('direction')

        directions = ['positive', 'negative']
        assert self.direction in directions, f"Expected one of {directions}, got {self.direction}"

    def score(
        self, observed_sum: float, expectations: np.array, penalty: float, q: float
    ):
        """
        Computes the score for the given q. (for the given records).

        The alternative hypothesis of MDSS assumes that there exists some constant multiplicative factor q > 1
        for the subset of records being scored by the scoring function.
        q is sometimes refered to as relative risk or severity.

        """
        raise NotImplementedError

    def dscore(self, observed_sum: float, expectations: np.array, q: float):
        """
        Computes the first derivative of the score function
        """
        raise NotImplementedError

    def q_dscore(self, observed_sum: float, expectations: np.array, q: float):
        """
        Computes the first derivative of the score function multiplied by the given q
        """
        raise NotImplementedError

    def qmle(self, observed_sum: float, expectations: np.array):
        """
        Computes the q which maximizes score (q_mle).
        """
        raise NotImplementedError

    def compute_qs(self, observed_sum: float, expectations: np.array, penalty: float):
        """
        Computes roots (qmin and qmax) of the score function (for the given records)
        """
        raise NotImplementedError



""" Tests for standard dataset classes """

from unittest.mock import patch
import numpy as np
import pandas as pd
import os

from aif360.datasets import AdultDataset
from aif360.datasets import BankDataset
from aif360.datasets import CompasDataset
from aif360.datasets import GermanDataset
from aif360.metrics import BinaryLabelDatasetMetric

pd.set_option('display.max_rows', 50)
pd.set_option('display.max_columns', 10)
pd.set_option('display.width', 200)

def test_compas():
    ''' Test default loading for compas '''
    # just test that there are no errors for default loading...
    compas_dataset = CompasDataset()
    compas_dataset.validate_dataset()

def test_german():
    ''' Test default loading for german '''
    german_dataset = GermanDataset()
    bldm = BinaryLabelDatasetMetric(german_dataset)
    assert bldm.num_instances() == 1000

def test_adult_test_set():
    ''' Test default loading for adult, test set '''
    adult_dataset = AdultDataset()
    test, _ = adult_dataset.split([15060])
    assert np.any(test.labels)

def test_adult():
    ''' Test default loading for adult, mean'''
    adult_dataset = AdultDataset()
    assert np.isclose(adult_dataset.labels.mean(), 0.2478, atol=5e-5)
    bldm = BinaryLabelDatasetMetric(adult_dataset)
    assert bldm.num_instances() == 45222

def test_adult_no_drop():
    ''' Test default loading for adult, number of instances '''
    adult_dataset = AdultDataset(protected_attribute_names=['sex'],
        privileged_classes=[['Male']], categorical_features=[],
        features_to_keep=['age', 'education-num'])
    bldm = BinaryLabelDatasetMetric(adult_dataset)
    assert bldm.num_instances() == 48842

def test_bank():
    ''' Test for errors during default loading '''
    bank_dataset = BankDataset()
    bank_dataset.validate_dataset()

def test_bank_priviliged_attributes():
    ''' Test if protected attribute age is correctly processed '''
    # Bank Data Set
    bank_dataset = BankDataset()
    num_priv = bank_dataset.protected_attributes.sum()
    
    # Raw data
    # TO DO: add file path. 
    filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            '..', 'aif360', 'data', 'raw', 'bank', 'bank-additional-full.csv')
    
    bank_dataset_unpreproc = pd.read_csv(filepath, sep = ";", na_values = ["unknown"])
    bank_dataset_unpreproc = bank_dataset_unpreproc.dropna()
    num_priv_raw = len(bank_dataset_unpreproc[(bank_dataset_unpreproc["age"] >= 25) & (bank_dataset_unpreproc["age"] < 60)])
    assert num_priv == num_priv_raw



    


from aif360.metrics import RegressionDatasetMetric
from aif360.datasets import RegressionDataset
import numpy as np
import pandas as pd

df = pd.DataFrame([
    ['r', 55],
    ['b', 65],
    ['b', 85],
    ['b', 70],
    ['r', 60],
    ['r', 50],
    ['r', 40],
    ['b', 30],
    ['r', 20],
    ['b', 10],
], columns=['s', 'score'])

dataset = RegressionDataset(df, dep_var_name='score', protected_attribute_names=['s'], privileged_classes=[['r']])
# sorted_dataset = RegressionDataset(df, dep_var_name='score', protected_attribute_names=['s'], privileged_classes=[['r']])


m = RegressionDatasetMetric(dataset=dataset,
                            privileged_groups=[{'s': 1}],
                            unprivileged_groups=[{'s': 0}])

def test_infeasible_index():
    actual = m.infeasible_index(target_prop={1: 0.5, 0: 0.5}, r=10)
    expected = (1, [3])
    assert actual == expected, f'Infeasible Index calculated wrong, got {actual}, expected {expected}'

def test_dcg():
    actual = m.discounted_cum_gain(normalized=False)
    expected = 2.6126967369231484
    assert abs(actual - expected) < 1e-6

def test_ndcg():
    actual = m.discounted_cum_gain(normalized=True, full_dataset=dataset)
    expected = 0.9205433036318259


from ..notebook_runner import notebook_run
import os


def test_demo_exp_grad_red_sklearn():
    nb, errors = notebook_run(os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '..', '..','examples', 'sklearn',
            'demo_exponentiated_gradient_reduction_sklearn.ipynb'))

    if len(errors) > 0:
        for err in errors:
            for tbi in err['traceback']:
                print(tbi)
        raise AssertionError("errors in notebook testcases")


import numpy as np

from aif360.datasets import StructuredDataset
from aif360.metrics import Metric, utils


class DatasetMetric(Metric):
    """Class for computing metrics based on one StructuredDataset."""

    def __init__(self, dataset, unprivileged_groups=None, privileged_groups=None):
        """
        Args:
            dataset (StructuredDataset): A StructuredDataset.
            privileged_groups (list(dict)): Privileged groups. Format is a list
                of `dicts` where the keys are `protected_attribute_names` and
                the values are values in `protected_attributes`. Each `dict`
                element describes a single group. See examples for more details.
            unprivileged_groups (list(dict)): Unprivileged groups in the same
                format as `privileged_groups`.

        Raises:
            TypeError: `dataset` must be a
                :obj:`~aif360.datasets.StructuredDataset` type.
            ValueError: `privileged_groups` and `unprivileged_groups` must be
                disjoint.

        Examples:
            >>> from aif360.datasets import GermanDataset
            >>> german = GermanDataset()
            >>> u = [{'sex': 1, 'age': 1}, {'sex': 0}]
            >>> p = [{'sex': 1, 'age': 0}]
            >>> dm = DatasetMetric(german, unprivileged_groups=u, privileged_groups=p)
        """
        if not isinstance(dataset, StructuredDataset):
            raise TypeError("'dataset' should be a StructuredDataset")

        # sets self.dataset
        super(DatasetMetric, self).__init__(dataset)

        # TODO: should this deepcopy?
        self.privileged_groups = privileged_groups
        self.unprivileged_groups = unprivileged_groups

        # don't check if nothing was provided
        if not self.privileged_groups or not self.unprivileged_groups:
            return

        priv_mask = utils.compute_boolean_conditioning_vector(
            self.dataset.protected_attributes,
            self.dataset.protected_attribute_names, self.privileged_groups)
        unpriv_mask = utils.compute_boolean_conditioning_vector(
            self.dataset.protected_attributes,
            self.dataset.protected_attribute_names, self.unprivileged_groups)
        if np.any(np.logical_and(priv_mask, unpriv_mask)):
            raise ValueError("'privileged_groups' and 'unprivileged_groups'"
                             " must be disjoint.")

    def _to_condition(self, privileged):
        """Converts a boolean condition to a group-specifying format that can be
        used to create a conditioning vector.
        """
        if privileged is True and self.privileged_groups is None:
            raise AttributeError("'privileged_groups' was not provided when "
                                 "this object was initialized.")
        if privileged is False and self.unprivileged_groups is None:
            raise AttributeError("'unprivileged_groups' was not provided when "
                                 "this object was initialized.")

        if privileged is None:
            return None
        return self.privileged_groups if privileged else self.unprivileged_groups

    def difference(self, metric_fun):
        """Compute difference of the metric for unprivileged and privileged
        groups.
        """
        return metric_fun(privileged=False) - metric_fun(privileged=True)

    def ratio(self, metric_fun):
        """Compute ratio of the metric for unprivileged and privileged groups.
        """
        return metric_fun(privileged=False) / metric_fun(privileged=True)

    def num_instances(self, privileged=None):
        """Compute the number of instances, :math:`n`, in the dataset conditioned
        on protected attributes if necessary.

        Args:
            privileged (bool, optional): Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.

        Raises:
            AttributeError: `privileged_groups` or `unprivileged_groups` must be
                must be provided at initialization to condition on them.
        """
        condition = self._to_condition(privileged)
        return utils.compute_num_instances(self.dataset.protected_attributes,
            self.dataset.instance_weights,
            self.dataset.protected_attribute_names, condition=condition)



"""
In-processing algorithms train a fair classifier (data in, predictions out).
"""
from aif360.sklearn.inprocessing.adversarial_debiasing import AdversarialDebiasing
from aif360.sklearn.inprocessing.exponentiated_gradient_reduction import ExponentiatedGradientReduction
from aif360.sklearn.inprocessing.grid_search_reduction import GridSearchReduction
from aif360.sklearn.inprocessing.infairness import SenSeI, SenSR

__all__ = [
    'AdversarialDebiasing',
    'ExponentiatedGradientReduction',
    'GridSearchReduction',
    'SenSeI', 'SenSR',
]




import numpy as np
from warnings import warn

from aif360.algorithms import Transformer
from aif360.metrics import utils
from aif360.metrics import BinaryLabelDatasetMetric, ClassificationMetric


class RejectOptionClassification(Transformer):

    """Reject option classification is a postprocessing technique that gives
    favorable outcomes to unpriviliged groups and unfavorable outcomes to
    priviliged groups in a confidence band around the decision boundary with the
    highest uncertainty [10]_.

    References:
        .. [10] F. Kamiran, A. Karim, and X. Zhang, "Decision Theory for
           Discrimination-Aware Classification," IEEE International Conference
           on Data Mining, 2012.
    """

    def __init__(self, unprivileged_groups, privileged_groups,
                low_class_thresh=0.01, high_class_thresh=0.99,
                num_class_thresh=100, num_ROC_margin=50,
                metric_name="Statistical parity difference",
                metric_ub=0.05, metric_lb=-0.05):
        """
        Args:
            unprivileged_groups (dict or list(dict)): Representation for
                unprivileged group.
            privileged_groups (dict or list(dict)): Representation for
                privileged group.
            low_class_thresh (float): Smallest classification threshold to use
                in the optimization. Should be between 0. and 1.
            high_class_thresh (float): Highest classification threshold to use
                in the optimization. Should be between 0. and 1.
            num_class_thresh (int): Number of classification thresholds between
                low_class_thresh and high_class_thresh for the optimization
                search. Should be > 0.
            num_ROC_margin (int): Number of relevant ROC margins to be used in
                the optimization search. Should be > 0.
            metric_name (str): Name of the metric to use for the optimization.
                Allowed options are "Statistical parity difference",
                "Average odds difference", "Equal opportunity difference".
            metric_ub (float): Upper bound of constraint on the metric value
            metric_lb (float): Lower bound of constraint on the metric value
        """
        super(RejectOptionClassification, self).__init__(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups,
            low_class_thresh=low_class_thresh, high_class_thresh=high_class_thresh,
            num_class_thresh=num_class_thresh, num_ROC_margin=num_ROC_margin,
            metric_name=metric_name)

        allowed_metrics = ["Statistical parity difference",
                           "Average odds difference",
                           "Equal opportunity difference"]

        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups

        self.low_class_thresh = low_class_thresh
        self.high_class_thresh = high_class_thresh
        self.num_class_thresh = num_class_thresh
        self.num_ROC_margin = num_ROC_margin
        self.metric_name = metric_name
        self.metric_ub = metric_ub
        self.metric_lb = metric_lb

        self.classification_threshold = None
        self.ROC_margin = None

        if ((self.low_class_thresh < 0.0) or (self.low_class_thresh > 1.0) or\
            (self.high_class_thresh < 0.0) or (self.high_class_thresh > 1.0) or\
            (self.low_class_thresh >= self.high_class_thresh) or\
            (self.num_class_thresh < 1) or (self.num_ROC_margin < 1)):

            raise ValueError("Input parameter values out of bounds")

        if metric_name not in allowed_metrics:
            raise ValueError("metric name not in the list of allowed metrics")

    def fit(self, dataset_true, dataset_pred):
        """Estimates the optimal classification threshold and margin for reject
        option classification that optimizes the metric provided.

        Note:
            The `fit` function is a no-op for this algorithm.

        Args:
            dataset_true (BinaryLabelDataset): Dataset containing the true
                `labels`.
            dataset_pred (BinaryLabelDataset): Dataset containing the predicted
                `scores`.

        Returns:
            RejectOptionClassification: Returns self.
        """

        fair_metric_arr = np.zeros(self.num_class_thresh*self.num_ROC_margin)
        balanced_acc_arr = np.zeros_like(fair_metric_arr)
        ROC_margin_arr = np.zeros_like(fair_metric_arr)
        class_thresh_arr = np.zeros_like(fair_metric_arr)

        cnt = 0
        # Iterate through class thresholds
        for class_thresh in np.linspace(self.low_class_thresh,
                                        self.high_class_thresh,
                                        self.num_class_thresh):

            self.classification_threshold = class_thresh
            if class_thresh <= 0.5:
                low_ROC_margin = 0.0
                high_ROC_margin = class_thresh
            else:
                low_ROC_margin = 0.0
                high_ROC_margin = (1.0-class_thresh)

            # Iterate through ROC margins
            for ROC_margin in np.linspace(
                                low_ROC_margin,
                                high_ROC_margin,
                                self.num_ROC_margin):
                self.ROC_margin = ROC_margin

                # Predict using the current threshold and margin
                dataset_transf_pred = self.predict(dataset_pred)

                dataset_transf_metric_pred = BinaryLabelDatasetMetric(
                                             dataset_transf_pred,
                                             unprivileged_groups=self.unprivileged_groups,
                                             privileged_groups=self.privileged_groups)
                classified_transf_metric = ClassificationMetric(
                                             dataset_true,
                                             dataset_transf_pred,
                                             unprivileged_groups=self.unprivileged_groups,
                                             privileged_groups=self.privileged_groups)

                ROC_margin_arr[cnt] = self.ROC_margin
                class_thresh_arr[cnt] = self.classification_threshold

                # Balanced accuracy and fairness metric computations
                balanced_acc_arr[cnt] = 0.5*(classified_transf_metric.true_positive_rate()\
                                       +classified_transf_metric.true_negative_rate())
                if self.metric_name == "Statistical parity difference":
                    fair_metric_arr[cnt] = dataset_transf_metric_pred.mean_difference()
                elif self.metric_name == "Average odds difference":
                    fair_metric_arr[cnt] = classified_transf_metric.average_odds_difference()
                elif self.metric_name == "Equal opportunity difference":
                    fair_metric_arr[cnt] = classified_transf_metric.equal_opportunity_difference()

                cnt += 1

        rel_inds = np.logical_and(fair_metric_arr >= self.metric_lb,
                                  fair_metric_arr <= self.metric_ub)
        if any(rel_inds):
            best_ind = np.where(balanced_acc_arr[rel_inds]
                                == np.max(balanced_acc_arr[rel_inds]))[0][0]
        else:
            warn("Unable to satisy fairness constraints")
            rel_inds = np.ones(len(fair_metric_arr), dtype=bool)
            best_ind = np.where(fair_metric_arr[rel_inds]
                                == np.min(fair_metric_arr[rel_inds]))[0][0]

        self.ROC_margin = ROC_margin_arr[rel_inds][best_ind]
        self.classification_threshold = class_thresh_arr[rel_inds][best_ind]

        return self

    def predict(self, dataset):
        """Obtain fair predictions using the ROC method.

        Args:
            dataset (BinaryLabelDataset): Dataset containing scores that will
                be used to compute predicted labels.

        Returns:
            dataset_pred (BinaryLabelDataset): Output dataset with potentially
            fair predictions obtain using the ROC method.
        """
        dataset_new = dataset.copy(deepcopy=False)

        fav_pred_inds = (dataset.scores > self.classification_threshold)
        unfav_pred_inds = ~fav_pred_inds

        y_pred = np.zeros(dataset.scores.shape)
        y_pred[fav_pred_inds] = dataset.favorable_label
        y_pred[unfav_pred_inds] = dataset.unfavorable_label

        # Indices of critical region around the classification boundary
        crit_region_inds = np.logical_and(
                dataset.scores <= self.classification_threshold+self.ROC_margin,
                dataset.scores > self.classification_threshold-self.ROC_margin)

        # Indices of privileged and unprivileged groups
        cond_priv = utils.compute_boolean_conditioning_vector(
                        dataset.protected_attributes,
                        dataset.protected_attribute_names,
                        self.privileged_groups)
        cond_unpriv = utils.compute_boolean_conditioning_vector(
                        dataset.protected_attributes,
                        dataset.protected_attribute_names,
                        self.unprivileged_groups)

        # New, fairer labels
        dataset_new.labels = y_pred
        dataset_new.labels[np.logical_and(crit_region_inds,
                            cond_priv.reshape(-1,1))] = dataset.unfavorable_label
        dataset_new.labels[np.logical_and(crit_region_inds,
                            cond_unpriv.reshape(-1,1))] = dataset.favorable_label

        return dataset_new

    def fit_predict(self, dataset_true, dataset_pred):
        """fit and predict methods sequentially."""
        return self.fit(dataset_true, dataset_pred).predict(dataset_pred)

# Function to obtain the pareto frontier
def _get_pareto_frontier(scores, return_mask = True):  # <- Fastest for many points
    """
    :param scores: An (n_points, n_scores) array
    :param return_mask: True to return a mask, False to return integer indices of efficient points.
    :return: An array of indices of pareto-efficient points.
        If return_mask is True, this will be an (n_points, ) boolean array
        Otherwise it will be a (n_efficient_points, ) integer array of indices.

    adapted from: https://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python
    """
    is_efficient = np.arange(scores.shape[0])
    n_points = scores.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for

    while next_point_index<len(scores):
        nondominated_point_mask = np.any(scores>=scores[next_point_index], axis=1)
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        scores = scores[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index])+1

    if return_mask:
        is_efficient_mask = np.zeros(n_points, dtype = bool)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask
    else:
        return is_efficient



from contextlib import nullcontext
from functools import partial
from inFairness import fairalgo, distances
import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from aif360.sklearn.inprocessing import SenSeI, SenSR
from aif360.sklearn.inprocessing.infairness import Dataset


def test_sensei_classification():
    """Tests whether SenSeI output matches original implementation."""
    X, y = make_classification(n_features=10)
    X = X.astype('float32')
    ds = Dataset(X, y)

    dx = distances.SVDSensitiveSubspaceDistance()
    dx.fit(X, 2)
    dy = distances.SquaredEuclideanDistance()
    dy.fit(2)

    torch.random.manual_seed(0)
    mlp = nn.Sequential(nn.Linear(10, 100), nn.ReLU(), nn.Linear(100, 2))
    aif360_sensei = SenSeI(
            mlp, criterion=nn.CrossEntropyLoss, distance_x=dx, distance_y=dy,
            rho=1., eps=0.1, auditor_nsteps=10, auditor_lr=0.01, max_epochs=5,
            optimizer=optim.Adam, lr=1e-3, predict_nonlinearity=None, verbose=0)
    y_pred = aif360_sensei.fit(X, y).predict_proba(X)
    assert aif360_sensei.regression_ == False

    torch.random.manual_seed(0)
    mlp = nn.Sequential(nn.Linear(10, 100), nn.ReLU(), nn.Linear(100, 2))
    orig_sensei = fairalgo.SenSeI(mlp, dx, dy, F.cross_entropy, 1., 0.1, 10, 0.01)
    orig_sensei.train()
    opt = optim.Adam(orig_sensei.parameters(), lr=1e-3)
    for _ in range(5):
        opt.zero_grad()
        batch = next(iter(torch.utils.data.DataLoader(ds, batch_size=len(ds))))
        loss = orig_sensei(*batch).loss
        loss.backward()
        opt.step()
    orig_sensei.eval()
    y_pred_orig = orig_sensei(torch.as_tensor(X), torch.as_tensor(y)).y_pred.detach().numpy()

    assert np.allclose(y_pred, y_pred_orig)

def test_sensr_regression():
    """Tests whether SenSR output matches original implementation."""
    X, y = make_regression(n_features=10)
    X, y = X.astype('float32'), y.astype('float32').reshape(-1, 1)
    ds = Dataset(X, y)

    dx = distances.MahalanobisDistances()
    dx.fit(torch.eye(10))

    torch.random.manual_seed(0)
    mlp = nn.Sequential(nn.Linear(10, 100), nn.ReLU(), nn.Linear(100, 1))
    aif360_sensei = SenSR(
            mlp, criterion=nn.MSELoss, distance_x=dx, eps=0.1, lr_lamb=1.,
            lr_param=1., auditor_nsteps=10, auditor_lr=0.01, max_epochs=5,
            optimizer=optim.Adam, lr=1e-3, predict_nonlinearity=None, verbose=0)
    y_pred = aif360_sensei.fit(X, y).predict(X)
    assert aif360_sensei.regression_ == True

    torch.random.manual_seed(0)
    mlp = nn.Sequential(nn.Linear(10, 100), nn.ReLU(), nn.Linear(100, 1))
    orig_sensei = fairalgo.SenSR(mlp, dx, F.mse_loss, 0.1, 1., 1., 10, 0.01)
    orig_sensei.train()
    opt = optim.Adam(orig_sensei.parameters(), lr=1e-3)
    for _ in range(5):
        opt.zero_grad()
        batch = next(iter(torch.utils.data.DataLoader(ds, batch_size=len(ds))))
        loss = orig_sensei(*batch).loss
        loss.backward()
        opt.step()
    orig_sensei.eval()
    y_pred_orig = orig_sensei(torch.as_tensor(X), torch.as_tensor(y)).y_pred.detach().numpy()

    assert np.allclose(y_pred, y_pred_orig)

@pytest.mark.parametrize(
    "y, criterion, raises", [
    # input unchanged
    ([1, 0, 1, 0], nn.CrossEntropyLoss, None),
    # input unchanged (wrong dtype)
    ([1., 0, 1, 0], nn.CrossEntropyLoss, RuntimeError),
    # input unchanged (must be 2D)
    ([[1.], [0], [1], [0]], nn.BCEWithLogitsLoss, None),
    # binarized to [[0], [1], [0], [1]]
    ([-1, 1, -1, 1], nn.BCEWithLogitsLoss, None),
    # binarized to [[0], [1], [0], [1]] (wrong shape)
    ([-1, 1, -1, 1], nn.CrossEntropyLoss, RuntimeError),
    # binarized to [[0], [1], [0], [1]]
    ([0, 2, 0, 2], nn.BCEWithLogitsLoss, None),
    # binarized to [[0], [1], [0], [1]] (wrong shape)
    ([0, 2, 0, 2], nn.CrossEntropyLoss, RuntimeError),
    # binarize to one-hot
    ([0, 1, 0, 2], nn.CrossEntropyLoss, None),
    # input unchanged -- detected regression (wrong shape)
    ([0.1, 1, 0, 2], nn.CrossEntropyLoss, RuntimeError),
    # input unchanged -- detected regression
    ([[0.1], [1], [0], [2]], nn.MSELoss, None),
    # binarized to [[0], [1], [0], [1]]
    (['a', 'b', 'a', 'b'], nn.BCEWithLogitsLoss, None),
    # input unchanged -- binary
    ([[0., 1], [1, 0], [0, 1], [1, 0]], nn.CrossEntropyLoss, None),
    # input unchanged -- multilabel
    ([[0., 1], [1, 0], [0, 1], [1, 0]], nn.BCEWithLogitsLoss, None),
    # multiclass-multioutput (not supported)
    ([[0, 1], [1, 2], [0, 1], [1, 0]], nn.CrossEntropyLoss, ValueError),
    # input unchanged -- detected multioutput regression
    ([[0.1, 1], [1, 0], [0, 1], [1, 0]], nn.MSELoss, None),
])
def test_target_encoding(y, criterion, raises):
    """Tests the automatic type casting for classification problems."""
    X = np.random.random((4, 2)).astype('float32')
    y = np.array(y)
    if criterion == nn.MSELoss:
        y = np.array(y, dtype='float32')
    classes = np.unique(y).tolist()
    if criterion == nn.BCEWithLogitsLoss or criterion == nn.MSELoss:
        ndim = 1 if y.ndim < 2 else y.shape[1]
    else:
        ndim = len(classes)
    dx = distances.SquaredEuclideanDistance()
    dx.fit(2)
    dy = distances.SquaredEuclideanDistance()
    dy.fit(ndim)
    mlp = nn.Sequential(nn.Linear(2, 10), nn.ReLU(), nn.Linear(10, ndim))
    sensei = SenSeI(mlp, criterion=criterion, distance_x=dx, distance_y=dy,
            rho=1., eps=0.1, auditor_nsteps=1, auditor_lr=0.01, max_epochs=1,
            optimizer=optim.Adam, lr=1e-3, verbose=0)
    with pytest.raises(raises) if raises is not None else nullcontext():
        sensei.fit(X, y)
        assert sensei.regression_ == (criterion == nn.MSELoss)
        if not sensei.regression_:
            assert sensei.classes_.tolist() == classes

dy = distances.SquaredEuclideanDistance()
dy.fit(2)
@pytest.mark.parametrize(
    "estimator_cls", [
    partial(SenSeI, distance_y=dy, rho=1., eps=0.1, auditor_nsteps=10, auditor_lr=0.01),
    partial(SenSR, eps=0.1, lr_lamb=1., lr_param=1., auditor_nsteps=10, auditor_lr=0.01),
])
def test_grid_pipe(estimator_cls):
    """Tests if SenSeI/SenSR works in a Pipeline and GridSearchCV."""
    X, y = make_classification(n_features=10)
    X = X.astype('float32')

    dx = distances.SquaredEuclideanDistance()
    dx.fit(10)

    mlp = nn.Sequential(nn.Linear(10, 100), nn.ReLU(), nn.Linear(100, 2))
    estimator = estimator_cls(mlp, criterion=nn.CrossEntropyLoss, distance_x=dx,
                              optimizer=optim.Adam, lr=1e-3, max_epochs=5, verbose=0)

    pipe = Pipeline([('scaler', StandardScaler()), ('estimator', estimator)])
    params = {'estimator__auditor_nsteps': [0, 10, 25]}
    grid = GridSearchCV(pipe, params, scoring='accuracy')
    grid.fit(X, y)
    assert not pd.DataFrame(grid.cv_results_).isna().any(axis=None)



# Copyright 2019 Seth V. Neel, Michael J. Kearns, Aaron L. Roth, Zhiwei Steven Wu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np

class RegOracle:
    """Class using regression oracle to solve CSC problem."""
    def __init__(self, b0, b1):
        self.b0 = b0
        self.b1 = b1

    def predict(self, X):
        """Predict labels on data set X."""
        c_0 = self.b0.predict(X)
        c_1 = self.b1.predict(X)
        y = (c_1 < c_0).astype('int')
        return y


class RandomLinearThresh:
    """Class random hyperplane classifier, used in experiments."""
    def __init__(self, d):
        self.coefficient = [np.random.uniform(-1, 1) for _ in range(d)]

    def predict(self, X):
        """Predict labels on data set X."""
        beta = self.coefficient
        n = X.shape[0]
        y = []
        for i in range(n):
            x_i = X.iloc[i, :]
            c_1 = np.dot(beta, x_i)
            y_i = int(c_1 < 0)
            y.append(y_i)
        return y


class LinearThresh:
    """Class hyperplane classifier."""
    def __init__(self, d):
        self.coefficient = d

    def predict(self, X):
        """Predict labels on data set X."""
        beta = self.coefficient
        n = X.shape[0]
        y = []
        for i in range(n):
            x_i = X.iloc[i, :]
            c_1 = np.dot(beta, x_i)
            y_i = int(c_1 < 0)
            y.append(y_i)
        return y



import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist

from aif360.datasets import StructuredDataset
from aif360.metrics import SampleDistortionMetric


data = np.arange(12).reshape((3, 4)).T
cols = ['one', 'two', 'three', 'label']
labs = np.ones((4, 1))

df = pd.DataFrame(data=np.concatenate((data, labs), axis=1), columns=cols)
sd = StructuredDataset(df=df, label_names=['label'],
    protected_attribute_names=['one', 'three'])

distorted = data + 1

sd_distorted = sd.copy(True)
sd_distorted.features = distorted

rand = np.random.randint(0, 10, (4, 4))
rand2 = np.random.randint(0, 10, (4, 3))
df_rand = pd.DataFrame(data=rand, columns=cols)
sd_rand = StructuredDataset(df=df_rand, label_names=['label'],
    protected_attribute_names=['one', 'three'])
sd_rand2 = sd_rand.copy(True)
sd_rand2.features = rand2


priv = [{'one': 1}]
unpriv = [{'one': 2}]

def test_euclidean_distance():
    sdm = SampleDistortionMetric(sd, sd_distorted)
    assert sdm.total_euclidean_distance() == 4*np.sqrt(3)

def test_manhattan_distance():
    sdm = SampleDistortionMetric(sd, sd_distorted)
    assert sdm.total_manhattan_distance() == 12

def test_mahalanobis_distance():
    sdm = SampleDistortionMetric(sd_rand, sd_rand2)
    assert np.isclose(sdm.total_mahalanobis_distance(),
        np.diag(cdist(rand[:, :3], rand2[:, :3], 'mahalanobis')).sum())

def test_conditional():
    sdm = SampleDistortionMetric(sd, sd_distorted, unprivileged_groups=unpriv,
        privileged_groups=priv)
    assert sdm.total_manhattan_distance(privileged=False) == 3

def test_average():
    sd_distorted.features[-1, -1] += 1
    sd.instance_weights = sd_distorted.instance_weights = np.array([1, 1, 1, 3])
    sdm = SampleDistortionMetric(sd, sd_distorted)
    assert sdm.average_manhattan_distance() == 3.5

def test_error():
    try:
        sd.protected_attributes -= 1
        sdm = SampleDistortionMetric(sd, sd_distorted)
    except ValueError:
        assert True



import numpy as np
import pandas as pd

from aif360.datasets import BinaryLabelDataset
from aif360.datasets.multiclass_label_dataset import MulticlassLabelDataset
from aif360.metrics import ClassificationMetric


def test_generalized_entropy_index():
    data = np.array([[0, 1],
                     [0, 0],
                     [1, 0],
                     [1, 1],
                     [1, 0],
                     [1, 0],
                     [2, 1],
                     [2, 0],
                     [2, 1],
                     [2, 1]])
    pred = data.copy()
    pred[[3, 9], -1] = 0
    pred[[4, 5], -1] = 1
    df = pd.DataFrame(data, columns=['feat', 'label'])
    df2 = pd.DataFrame(pred, columns=['feat', 'label'])
    bld = BinaryLabelDataset(df=df, label_names=['label'],
        protected_attribute_names=['feat'])
    bld2 = BinaryLabelDataset(df=df2, label_names=['label'],
        protected_attribute_names=['feat'])
    cm = ClassificationMetric(bld, bld2)

    assert cm.generalized_entropy_index() == 0.2

    pred = data.copy()
    pred[:, -1] = np.array([0, 1, 1, 0, 0, 0, 0, 1, 1, 1])
    df2 = pd.DataFrame(pred, columns=['feat', 'label'])
    bld2 = BinaryLabelDataset(df=df2, label_names=['label'],
        protected_attribute_names=['feat'])
    cm = ClassificationMetric(bld, bld2)

    assert cm.generalized_entropy_index() == 0.3

def test_theil_index():
    data = np.array([[0, 1],
                     [0, 0],
                     [1, 0],
                     [1, 1],
                     [1, 0],
                     [1, 0],
                     [2, 1],
                     [2, 0],
                     [2, 1],
                     [2, 1]])
    pred = data.copy()
    pred[[3, 9], -1] = 0
    pred[[4, 5], -1] = 1
    df = pd.DataFrame(data, columns=['feat', 'label'])
    df2 = pd.DataFrame(pred, columns=['feat', 'label'])
    bld = BinaryLabelDataset(df=df, label_names=['label'],
        protected_attribute_names=['feat'])
    bld2 = BinaryLabelDataset(df=df2, label_names=['label'],
        protected_attribute_names=['feat'])
    cm = ClassificationMetric(bld, bld2)

    assert cm.theil_index() == 4*np.log(2)/10

def test_between_all_groups():
    data = np.array([[0, 1],
                     [0, 0],
                     [1, 0],
                     [1, 1],
                     [1, 0],
                     [1, 0],
                     [2, 1],
                     [2, 0],
                     [2, 1],
                     [2, 1]])
    pred = data.copy()
    pred[[3, 9], -1] = 0
    pred[[4, 5], -1] = 1
    df = pd.DataFrame(data, columns=['feat', 'label'])
    df2 = pd.DataFrame(pred, columns=['feat', 'label'])
    bld = BinaryLabelDataset(df=df, label_names=['label'],
        protected_attribute_names=['feat'])
    bld2 = BinaryLabelDataset(df=df2, label_names=['label'],
        protected_attribute_names=['feat'])
    cm = ClassificationMetric(bld, bld2)

    b = np.array([1, 1, 1.25, 1.25, 1.25, 1.25, 0.75, 0.75, 0.75, 0.75])
    assert cm.between_all_groups_generalized_entropy_index() == 1/20*np.sum(b**2 - 1)

def test_between_group():
    data = np.array([[0, 0, 1],
                     [0, 1, 0],
                     [1, 1, 0],
                     [1, 1, 1],
                     [1, 0, 0],
                     [1, 0, 0]])
    pred = data.copy()
    pred[[0, 3], -1] = 0
    pred[[4, 5], -1] = 1
    df = pd.DataFrame(data, columns=['feat', 'feat2', 'label'])
    df2 = pd.DataFrame(pred, columns=['feat', 'feat2', 'label'])
    bld = BinaryLabelDataset(df=df, label_names=['label'],
        protected_attribute_names=['feat', 'feat2'])
    bld2 = BinaryLabelDataset(df=df2, label_names=['label'],
        protected_attribute_names=['feat', 'feat2'])
    cm = ClassificationMetric(bld, bld2, unprivileged_groups=[{'feat': 0}],
        privileged_groups=[{'feat': 1}])

    b = np.array([0.5, 0.5, 1.25, 1.25, 1.25, 1.25])
    assert cm.between_group_generalized_entropy_index() == 1/12*np.sum(b**2 - 1)
    
def test_multiclass_confusion_matrix():
    data = np.array([[0, 1],
                     [0, 0],
                     [1, 0],
                     [1, 1],
                     [1, 0],
                     [1, 2],
                     [2, 1],
                     [2, 0],
                     [2, 2],
                     [2, 1]])
    pred = data.copy()
    pred[3,1] = 0
    pred[4,1] = 2
    
    df = pd.DataFrame(data, columns=['feat', 'label'])
    df2 = pd.DataFrame(pred, columns=['feat', 'label'])

    favorable_values = [0,1]
    unfavorable_values = [2]
    mcld = MulticlassLabelDataset(favorable_label = favorable_values, unfavorable_label = unfavorable_values , df = df , label_names=['label'],
        protected_attribute_names=['feat'])
    mcld2 = MulticlassLabelDataset(favorable_label = favorable_values, unfavorable_label = unfavorable_values , df=df2, label_names=['label'],
        protected_attribute_names=['feat'])
    cm = ClassificationMetric(mcld, mcld2, unprivileged_groups=[{'feat': 2}],
        privileged_groups=[{'feat': 0},{'feat': 1}])
    confusion_matrix = cm.binary_confusion_matrix()
    
    actual_labels_df = df[['label']].values
    actual_labels_df2 = df2[['label']].values
   
    assert np.all(actual_labels_df == mcld.labels)
    assert np.all(actual_labels_df2 == mcld2.labels)
   
    assert confusion_matrix == {'TP': 7.0, 'FN': 1.0, 'TN': 2.0, 'FP': 0.0}

    fnr = cm.false_negative_rate_difference()
    assert fnr == -0.2
    
def test_generalized_binary_confusion_matrix():
    data = np.array([[0, 1],
                     [0, 0],
                     [1, 0],
                     [1, 1],
                     [1, 0],
                     [1, 0],
                     [1, 2],
                     [0, 0],
                     [0, 0],
                     [1, 2]])
   
    pred = np.array([[0, 1, 0.8],
                     [0, 0, 0.6],
                     [1, 0, 0.7],
                     [1, 1, 0.8],
                     [1, 2, 0.36],
                     [1, 0, 0.82],
                     [1, 1, 0.79],
                     [0, 2, 0.42],
                     [0, 1, 0.81],
                     [1, 2, 0.3]])
    df = pd.DataFrame(data, columns=['feat', 'label'])
    df2 = pd.DataFrame(pred, columns=['feat', 'label', 'score'])

    favorable_values = [0,1]
    unfavorable_values = [2]
    
    mcld = MulticlassLabelDataset(df=df, label_names=['label'],
        protected_attribute_names=['feat'],favorable_label = favorable_values, unfavorable_label = unfavorable_values)
    
    mcld2 = MulticlassLabelDataset(df=df2, label_names=['label'], scores_names=['score'],
        protected_attribute_names=['feat'],favorable_label = favorable_values, unfavorable_label = unfavorable_values)
    
    
    cm = ClassificationMetric(mcld, mcld2, unprivileged_groups=[{'feat': 0}],
        privileged_groups=[{'feat': 1}])

    gen_confusion_matrix = cm.generalized_binary_confusion_matrix()
 
    gtp = cm.num_generalized_true_positives()
    assert round(gtp,2) == 5.31
    gfp = cm.num_generalized_false_positives()
    assert gfp == 1.09
   
   


from aif360.detectors.mdss.ScoringFunctions.ScoringFunction import ScoringFunction
from aif360.detectors.mdss.ScoringFunctions import optim

import numpy as np


class Bernoulli(ScoringFunction):
    def __init__(self, **kwargs):
        """
        Bernoulli score function. May be appropriate to use when the outcome of
        interest is assumed to be Bernoulli distributed or Binary.

        kwargs must contain
        'direction (str)' - direction of the severity; could be higher than expected outcomes ('positive') or lower than expected ('negative')
        """

        super(Bernoulli, self).__init__(**kwargs)

    def score(self, observed_sum: float, expectations: np.array, penalty: float, q: float):
        """
        Computes bernoulli bias score for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty term. Should be positive
        :param q: current value of q
        :return: bias score for the current value of q
        """

        assert q > 0, (
            "Warning: calling compute_score_given_q with "
            "observed_sum=%.2f, expectations of length=%d, penalty=%.2f, q=%.2f"
            % (observed_sum, len(expectations), penalty, q)
        )

        ans = observed_sum * np.log(q) - np.log(1 - expectations + q * expectations).sum() - penalty
        return ans

    def qmle(self, observed_sum: float, expectations: np.array):
        """
        Computes the q which maximizes score (q_mle).

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        """
        direction = self.direction        
        ans = optim.bisection_q_mle(self, observed_sum, expectations, direction=direction)
        return ans

    def compute_qs(self, observed_sum: float, expectations: np.array, penalty: float):
        """
        Computes roots (qmin and qmax) of the score function for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty coefficient
        """
        direction = self.direction
        q_mle = self.qmle(observed_sum, expectations)

        if self.score(observed_sum, expectations, penalty, q_mle) > 0:
            exist = 1
            q_min = optim.bisection_q_min(self, observed_sum, expectations, penalty, q_mle)
            q_max = optim.bisection_q_max(self, observed_sum, expectations, penalty, q_mle)
        else:
            # there are no roots
            exist = 0
            q_min = 0
            q_max = 0

        # only consider the desired direction, positive or negative
        if exist:
            exist, q_min, q_max = optim.direction_assertions(direction, q_min, q_max)

        ans = [exist, q_mle, q_min, q_max]
        return ans

    def q_dscore(self, observed_sum:float, expectations:np.array, q:float):
        """
        This actually computes q times the slope, which has the same sign as the slope since q is positive.
        score = Y log q - \sum_i log(1-p_i+qp_i)
        dscore/dq = Y/q - \sum_i (p_i/(1-p_i+qp_i))
        q dscore/dq = Y - \sum_i (qp_i/(1-p_i+qp_i))

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param q: current value of q
        :return: q dscore/dq
        """
        ans = observed_sum - (q * expectations / (1 - expectations + q * expectations)).sum()
        return ans



# Copyright 2019 Seth V. Neel, Michael J. Kearns, Aaron L. Roth, Zhiwei Steven Wu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from matplotlib import pyplot as plt


def plot_single(errors_t, fp_diff_t, max_iters, gamma, C):
    """Plot the errors and false positive rate disparity over time.

    :param errors_t: list of errors at each iteration
    :param fp_diff_t: list of fp rate disparity at each iteration
    :param max_iters: time horizon T of the algorithm
    :param gamma: input gamma disparity
    :param C: input C parameter - see gerryfair_classifier.py
    """
    # plot errors
    x = range(max_iters - 1)
    y_t = errors_t
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    ax1.plot(x, y_t)
    plt.ylabel('average error of mixture')
    plt.xlabel('iterations')
    plt.title('error vs. time: C: {}, gamma: {}'.format(C, gamma))
    plt.show()

    # plot fp disparity
    x = range(max_iters - 1)
    y_t = fp_diff_t
    fig2 = plt.figure()
    ax2 = fig2.add_subplot(111)
    ax2.plot(x, y_t)
    plt.ylabel('fp_diff*group_size')
    plt.xlabel('iterations')
    plt.title('fp_diff*size vs. time: C: {}, gamma: {}'.format(C, gamma))
    ax2.plot(x, [gamma] * len(y_t))
    plt.show()


import os

import pandas as pd

from aif360.datasets import StandardDataset


default_mappings = {
    'label_maps': [{1.0: 'Good Credit', 2.0: 'Bad Credit'}],
    'protected_attribute_maps': [{1.0: 'Male', 0.0: 'Female'},
                                 {1.0: 'Old', 0.0: 'Young'}],
}

def default_preprocessing(df):
    """Adds a derived sex attribute based on personal_status."""
    # TODO: ignores the value of privileged_classes for 'sex'
    status_map = {'A91': 'male', 'A93': 'male', 'A94': 'male',
                  'A92': 'female', 'A95': 'female'}
    df['sex'] = df['personal_status'].replace(status_map)

    return df

class GermanDataset(StandardDataset):
    """German credit Dataset.

    See :file:`aif360/data/raw/german/README.md`.
    """

    def __init__(self, label_name='credit', favorable_classes=[1],
                 protected_attribute_names=['sex', 'age'],
                 privileged_classes=[['male'], lambda x: x > 25],
                 instance_weights_name=None,
                 categorical_features=['status', 'credit_history', 'purpose',
                     'savings', 'employment', 'other_debtors', 'property',
                     'installment_plans', 'housing', 'skill_level', 'telephone',
                     'foreign_worker'],
                 features_to_keep=[], features_to_drop=['personal_status'],
                 na_values=[], custom_preprocessing=default_preprocessing,
                 metadata=default_mappings):
        """See :obj:`StandardDataset` for a description of the arguments.

        By default, this code converts the 'age' attribute to a binary value
        where privileged is `age > 25` and unprivileged is `age <= 25` as
        proposed by Kamiran and Calders [1]_.

        References:
            .. [1] F. Kamiran and T. Calders, "Classifying without
               discriminating," 2nd International Conference on Computer,
               Control and Communication, 2009.

        Examples:
            In some cases, it may be useful to keep track of a mapping from
            `float -> str` for protected attributes and/or labels. If our use
            case differs from the default, we can modify the mapping stored in
            `metadata`:

            >>> label_map = {1.0: 'Good Credit', 0.0: 'Bad Credit'}
            >>> protected_attribute_maps = [{1.0: 'Male', 0.0: 'Female'}]
            >>> gd = GermanDataset(protected_attribute_names=['sex'],
            ... privileged_classes=[['male']], metadata={'label_map': label_map,
            ... 'protected_attribute_maps': protected_attribute_maps})

            Now this information will stay attached to the dataset and can be
            used for more descriptive visualizations.
        """

        filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                '..', 'data', 'raw', 'german', 'german.data')
        # as given by german.doc
        column_names = ['status', 'month', 'credit_history',
            'purpose', 'credit_amount', 'savings', 'employment',
            'investment_as_income_percentage', 'personal_status',
            'other_debtors', 'residence_since', 'property', 'age',
            'installment_plans', 'housing', 'number_of_credits',
            'skill_level', 'people_liable_for', 'telephone',
            'foreign_worker', 'credit']
        try:
            df = pd.read_csv(filepath, sep=' ', header=None, names=column_names,
                             na_values=na_values)
        except IOError as err:
            print("IOError: {}".format(err))
            print("To use this class, please download the following files:")
            print("\n\thttps://archive.ics.uci.edu/ml/machine-learning-databases/statlog/german/german.data")
            print("\thttps://archive.ics.uci.edu/ml/machine-learning-databases/statlog/german/german.doc")
            print("\nand place them, as-is, in the folder:")
            print("\n\t{}\n".format(os.path.abspath(os.path.join(
                os.path.abspath(__file__), '..', '..', 'data', 'raw', 'german'))))
            import sys
            sys.exit(1)

        super(GermanDataset, self).__init__(df=df, label_name=label_name,
            favorable_classes=favorable_classes,
            protected_attribute_names=protected_attribute_names,
            privileged_classes=privileged_classes,
            instance_weights_name=instance_weights_name,
            categorical_features=categorical_features,
            features_to_keep=features_to_keep,
            features_to_drop=features_to_drop, na_values=na_values,
            custom_preprocessing=custom_preprocessing, metadata=metadata)



import numpy as np

from aif360.algorithms.inprocessing.celisMeta.General import General


class StatisticalRate(General):
    def getExpectedGrad(self, dist, a, b, params, samples, mu, z_prior):
        l_1, l_2 = params

        t, c_1, c_2 = self.getValueForX(dist, a, b, params, z_prior, samples,
                                        return_cs=True)

        t1 = t * c_1/np.sqrt(t**2 + mu**2)
        t2 = t * c_2/np.sqrt(t**2 + mu**2)

        exp1 = np.mean(t1)
        exp2 = np.mean(t2)
        dl1 = exp1 - b + (b-a)/2 + (b-a)*l_1 / (2*np.sqrt(l_1**2 + mu**2))
        dl2 = exp2 - b + (b-a)/2 + (b-a)*l_2 / (2*np.sqrt(l_2**2 + mu**2))
        return dl1, dl2

    def getValueForX(self, dist, a, b, params, z_prior, x, return_cs=False):
        l_1, l_2 = params
        z_0, z_1 = 1-z_prior, z_prior

        pos = np.ones(len(x))
        prob_1_1 = self.prob(dist, np.c_[x, pos, pos])
        prob_m1_1 = self.prob(dist, np.c_[x, -pos, pos])
        prob_1_0 = self.prob(dist, np.c_[x, pos, np.zeros(len(x))])
        prob_m1_0 = self.prob(dist, np.c_[x, -pos, np.zeros(len(x))])

        total = prob_1_1 + prob_1_0 + prob_m1_0 + prob_m1_1
        # if total == 0:
        #     return 0

        prob_y_1 = (prob_1_1 + prob_1_0) / total
        prob_z_0 = (prob_m1_0 + prob_1_0) / total
        prob_z_1 = (prob_m1_1 + prob_1_1) / total

        c_0 = prob_y_1 - 0.5
        c_1 = prob_z_0 / z_0
        c_2 = prob_z_1 / z_1

        t = c_0 + c_1*l_1 + c_2*l_2
        if return_cs:
            return t, c_1, c_2
        else:
            return t

    def getFuncValue(self, dist, a, b, params, samples, z_prior):
        l_1, l_2 = params

        exp = np.mean(np.abs(self.getValueForX(dist, a, b, params, z_prior,
                                               samples)))
        result = exp - b*l_1 - b*l_2
        if l_1 > 0:
            result += (b-a)*l_1
        if l_2 > 0:
            result += (b-a)*l_2

        return result

    @property
    def num_params(self):
        return 2

    def init_params(self, i):
        return [i-5] * self.num_params

    def gamma(self, y_true, y_pred, sens):
        pos_0 = np.mean(y_pred[sens == 0] == 1)
        pos_1 = np.mean(y_pred[sens == 1] == 1)
        if pos_0 == 0 or pos_1 == 0:
            return 0
        return min(pos_0/pos_1, pos_1/pos_0)



from io import BytesIO
import os
import urllib

import pandas as pd
from sklearn.model_selection import train_test_split

from aif360.sklearn.datasets.utils import standardize_dataset, Dataset


# cache location
DATA_HOME_DEFAULT = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 '..', 'data', 'raw')
LSAC_URL = "https://github.com/jkomiyama/fairregresion/raw/master/dataset/law/lsac.sas7bdat"

def fetch_lawschool_gpa(subset="all", *, data_home=None, cache=True,
                        binary_race=True, fillna_gender="female",
                        usecols=["race", "gender", "lsat", "ugpa"],
                        dropcols=None, numeric_only=False, dropna=True):
    """Load the Law School GPA dataset.

    Optionally binarizes 'race' to 'white' (privileged) or 'black' (unprivileged).
    The other protected attribute is gender ('male' is privileged and 'female'
    is unprivileged). The outcome variable is standardized first year GPA
    ('zfygpa'). Note: this is a continuous variable, i.e., a regression task.

    Args:
        subset ({'train', 'test', or 'all'}, optional): Select the dataset to
            load: 'train' for the training set, 'test' for the test set, 'all'
            for both.
        data_home (string, optional): Specify another download and cache folder
            for the datasets. By default all AIF360 datasets are stored in
            'aif360/sklearn/data/raw' subfolders.
        cache (bool): Whether to cache downloaded datasets.
        binary_race (bool, optional): Filter only white and black students.
        fillna_gender (str or None, optional): Fill NA values for gender with
            this value. If `None`, leave as NA. Note: this is used for backward-
            compatibility with tempeh and may be dropped in later versions.
        usecols (single label or list-like, optional): Feature column(s) to
            keep. All others are dropped.
        dropcols (single label or list-like, optional): Feature column(s) to
            drop.
        numeric_only (bool): Drop all non-numeric feature columns.
        dropna (bool): Drop rows with NAs.

    Returns:
        namedtuple: Tuple containing X, y, and sample_weights for the Law School
        GPA dataset accessible by index or name.
    """
    if subset not in {'train', 'test', 'all'}:
        raise ValueError("subset must be either 'train', 'test', or 'all'; "
                         "cannot be {}".format(subset))

    cache_path = os.path.join(data_home or DATA_HOME_DEFAULT,
                              os.path.basename(LSAC_URL))
    if cache and os.path.isfile(cache_path):
        df = pd.read_sas(cache_path, encoding="utf-8")
    else:
        data = urllib.request.urlopen(LSAC_URL).read()
        if cache:
            os.makedirs(os.path.dirname (cache_path), exist_ok=True)
            with open(cache_path, 'wb') as f:
                f.write(data)
        df = pd.read_sas(BytesIO(data), format="sas7bdat", encoding="utf-8")

    df.race = df.race1.astype('category')
    if binary_race:
        df.race = df.race.cat.set_categories(['black', 'white'], ordered=True)

    # for backwards-compatibility with tempeh
    if fillna_gender is not None:
        df.gender = df.gender.fillna(fillna_gender)
    df.gender = df.gender.astype('category').cat.set_categories(
        ['female', 'male'], ordered=True)

    ds = standardize_dataset(df, prot_attr=['race', 'gender'], target='zfygpa',
                               usecols=usecols, dropcols=dropcols,
                               numeric_only=numeric_only, dropna=dropna)

    # for backwards-compatibility with tempeh
    train_X, test_X, train_y, test_y = train_test_split(*ds, test_size=0.33, random_state=123)
    if subset == "train":
        return Dataset(train_X, train_y)
    elif subset == "test":
        return Dataset(test_X, test_y)
    else:
        X = pd.concat([train_X, test_X], axis=0)
        y = pd.concat([train_y, test_y], axis=0)
        return Dataset(X, y)
